# Copyright (C) 2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

import logging
import sys

from xl.nls import gettext as _
from xl.playlist import Playlist, is_valid_playlist, import_playlist
from xl import common, event, main, player, providers, settings, trax, xdg

from xlgui.widgets.common import AutoScrollTreeView
from xlgui.widgets.notebook import NotebookPage
from xlgui.widgets import dialogs, menu, menuitems, playlist_columns
from xlgui import guiutil, icons

logger = logging.getLogger(__name__)


def default_get_playlist_func(parent, context):
    return player.QUEUE.current_playlist


class ModesMenuItem(menu.MenuItem):
    """
    A menu item having a submenu containing entries for shuffle modes.

    Defaults to adjusting the currently-playing playlist.
    """

    modetype = ''
    display_name = ""

    def __init__(self, name, after, get_playlist_func=default_get_playlist_func):
        menu.MenuItem.__init__(self, name, None, after)
        self.get_playlist_func = get_playlist_func

    def factory(self, menu, parent, context):
        item = Gtk.ImageMenuItem.new_with_mnemonic(self.display_name)
        image = Gtk.Image.new_from_icon_name(
            'media-playlist-' + self.modetype, size=Gtk.IconSize.MENU
        )
        item.set_image(image)
        submenu = self.create_mode_submenu(item)
        item.set_submenu(submenu)
        pl = self.get_playlist_func(parent, context)
        item.set_sensitive(pl is not None)
        return item

    def create_mode_submenu(self, parent_item):
        names = getattr(Playlist, "%s_modes" % self.modetype)
        displays = getattr(Playlist, "%s_mode_names" % self.modetype)
        items = []
        previous = None
        for name, display in zip(names, displays):
            after = [previous] if previous else []
            item = menu.radio_menu_item(
                name,
                after,
                display,
                '%s_modes' % self.modetype,
                self.mode_is_selected,
                self.on_mode_activated,
            )
            items.append(item)
            if previous is None:
                items.append(menu.simple_separator("sep", [items[-1].name]))
            previous = items[-1].name
        m = menu.Menu(parent_item)
        for item in items:
            m.add_item(item)
        return m

    def mode_is_selected(self, name, parent, context):
        pl = self.get_playlist_func(parent, context)
        if pl is None:
            return False
        return getattr(pl, "%s_mode" % self.modetype) == name

    def on_mode_activated(self, widget, name, parent, context):
        pl = self.get_playlist_func(parent, context)
        if pl is None:
            return False
        setattr(pl, "%s_mode" % self.modetype, name)


class ShuffleModesMenuItem(ModesMenuItem):
    modetype = 'shuffle'
    display_name = _("S_huffle")


class RepeatModesMenuItem(ModesMenuItem):
    modetype = 'repeat'
    display_name = _("R_epeat")


class DynamicModesMenuItem(ModesMenuItem):
    modetype = 'dynamic'
    display_name = _("_Dynamic")


class RemoveCurrentMenuItem(menu.MenuItem):
    """
    Allows for removing the currently playing
    track from the current playlist
    """

    def __init__(self, after, get_playlist_func=default_get_playlist_func):
        menu.MenuItem.__init__(self, 'remove-current', None, after)
        self.get_playlist_func = get_playlist_func

    def factory(self, menu, parent, context):
        """
        Sets up the menu item
        """
        item = Gtk.ImageMenuItem.new_with_mnemonic(
            _('Remove _Current Track From Playlist')
        )
        item.set_image(Gtk.Image.new_from_icon_name('list-remove', Gtk.IconSize.MENU))
        item.connect('activate', self.on_activate, parent, context)

        if player.PLAYER.is_stopped():
            item.set_sensitive(False)

        return item

    def on_activate(self, menuitem, parent, context):
        """
        Removes the currently playing track from the current playlist
        """
        playlist = self.get_playlist_func(parent, context)

        if playlist and playlist.current == player.PLAYER.current:
            del playlist[playlist.current_position]


class RandomizeMenuItem(menu.MenuItem):
    """
    A menu item which randomizes the full
    playlist or the current selection
    """

    def __init__(self, after):
        menu.MenuItem.__init__(self, 'randomize', None, after)

    def factory(self, menu, parent, context):
        """
        Sets up the menu item
        """
        label = _('R_andomize Playlist')

        if not context['selection-empty']:
            label = _('R_andomize Selection')

        item = Gtk.MenuItem.new_with_mnemonic(label)
        item.connect('activate', self.on_activate, parent, context)

        return item

    def on_activate(self, menuitem, parent, context):
        """
        Randomizes the playlist or the selection
        """
        positions = [path[0] for path in context['selected-paths']]
        # Randomize the full playlist if only one track was selected
        positions = positions if len(positions) > 1 else []

        # If you don't unselect before randomizing, a large number of selected will
        # kill performance
        parent.get_selection().unselect_all()
        with parent.handler_block(parent._cursor_changed):
            context['playlist'].randomize(positions)


# do this in a function to avoid polluting the global namespace
def __create_playlist_tab_context_menu():
    smi = menu.simple_menu_item
    sep = menu.simple_separator
    items = []
    items.append(
        smi(
            'new-tab',
            [],
            _("_New Playlist"),
            'tab-new',
            lambda w, n, o, c: o.tab.notebook.create_new_playlist(),
        )
    )
    items.append(sep('new-tab-sep', ['new-tab']))

    items.append(
        smi(
            'save',
            ['new-tab-sep'],
            _("_Save"),
            'document-save',
            callback=lambda w, n, p, c: p.on_save(),
            condition_fn=lambda n, p, c: (
                p.can_save()
                and main.exaile().playlists.has_playlist_name(p.playlist.name)
            ),
        )
    )
    items.append(
        smi(
            'saveas',
            ['save'],
            _("Save _As"),
            'document-save-as',
            callback=lambda w, n, p, c: p.on_saveas(),
            condition_fn=lambda n, p, c: p.can_saveas(),
        )
    )
    items.append(
        smi(
            'rename',
            ['saveas'],
            _("_Rename"),
            None,
            callback=lambda w, n, p, c: p.tab.start_rename(),
            condition_fn=lambda n, p, c: p.tab.can_rename(),
        )
    )
    items.append(
        smi(
            'clear',
            ['rename'],
            _("_Clear"),
            'edit-clear-all',
            lambda w, n, o, c: o.playlist.clear(),
        )
    )
    items.append(sep('tab-close-sep', ['clear']))

    def _get_pl_func(o, c):
        return o.playlist

    items.append(
        menuitems.ExportPlaylistMenuItem('export', ['tab-close-sep'], _get_pl_func)
    )
    items.append(
        menuitems.ExportPlaylistFilesMenuItem('export-files', ['export'], _get_pl_func)
    )
    items.append(sep('tab-export-sep', ['export']))
    items.append(
        smi(
            'tab-close',
            ['tab-export-sep'],
            _("Close _Tab"),
            'window-close',
            lambda w, n, o, c: o.tab.close(),
        )
    )
    for item in items:
        providers.register('playlist-tab-context-menu', item)


__create_playlist_tab_context_menu()


class PlaylistContextMenu(menu.ProviderMenu):
    def __init__(self, page):
        """
        :param page: The :class:`PlaylistPage` this menu is
            associated with.
        """
        menu.ProviderMenu.__init__(self, 'playlist-context-menu', page)
        self.attach_to_widget(page, None)

    def get_context(self):
        context = common.LazyDict(self._parent)
        context['playlist'] = lambda name, parent: parent.playlist
        context['selection-empty'] = (
            lambda name, parent: parent.get_selection_count() == 0
        )
        context['selected-paths'] = lambda name, parent: parent.get_selected_paths()
        context['selected-items'] = lambda name, parent: parent.get_selected_items()
        context['selected-tracks'] = lambda name, parent: parent.get_selected_tracks()
        context['selection-count'] = lambda name, parent: parent.get_selection_count()
        return context


class SPATMenuItem(menu.MenuItem):
    """
    Menu item allowing for toggling playback
    stop after the selected track (SPAT)
    """

    def __init__(self, name, after):
        menu.MenuItem.__init__(self, name, None, after)

    def factory(self, menu, parent, context):
        """
        Generates the menu item
        """
        display_name = _('_Stop Playback After This Track')
        icon_name = 'media-playback-stop'

        if context['selected-items']:
            selection_position = context['selected-items'][0][0]

            if selection_position == parent.playlist.spat_position:
                display_name = _('_Continue Playback After This Track')
                icon_name = 'media-playback-start'

        menuitem = Gtk.ImageMenuItem.new_with_mnemonic(display_name)
        menuitem.set_image(Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.MENU))
        menuitem.connect('activate', self.on_menuitem_activate, parent, context)

        return menuitem

    def on_menuitem_activate(self, menuitem, parent, context):
        """
        Toggles the SPAT state
        """
        selection_position = context['selected-items'][0][0]

        if selection_position == parent.playlist.spat_position:
            parent.playlist.spat_position = -1
        else:
            parent.playlist.spat_position = selection_position


def __create_playlist_context_menu():
    smi = menu.simple_menu_item
    sep = menu.simple_separator
    items = []

    items.append(menuitems.EnqueueMenuItem('enqueue', []))

    items.append(SPATMenuItem('toggle-spat', [items[-1].name]))

    def rating_get_tracks_func(menuobj, parent, context):
        return [row[1] for row in context['selected-items']]

    items.append(menuitems.RatingMenuItem('rating', [items[-1].name]))

    # TODO: custom playlist item here
    items.append(sep('sep1', [items[-1].name]))

    def remove_tracks_cb(widget, name, playlistpage, context):
        tracks = context['selected-items']
        playlist = playlistpage.playlist
        # If it's all one block, just delete it in one chunk for
        # maximum speed.
        positions = [t[0] for t in tracks]
        if positions == list(range(positions[0], positions[0] + len(positions))):
            del playlist[positions[0] : positions[0] + len(positions)]
        else:
            for position, track in tracks[::-1]:
                del playlist[position]

    items.append(
        smi(
            'remove',
            [items[-1].name],
            _("_Remove from Playlist"),
            'list-remove',
            remove_tracks_cb,
        )
    )

    items.append(RandomizeMenuItem([items[-1].name]))

    def playlist_menu_condition(name, parent, context):
        """
        Returns True if the containing notebook's tab bar is hidden
        """
        scrolledwindow = parent.get_parent()
        page = scrolledwindow.get_parent()
        return not page.tab.notebook.get_show_tabs()

    items.append(
        smi(
            'playlist-menu',
            [items[-1].name],
            _('Playlist'),
            submenu=menu.ProviderMenu('playlist-tab-context-menu', None),
            condition_fn=playlist_menu_condition,
        )
    )

    items.append(sep('sep2', [items[-1].name]))

    items.append(
        smi(
            'properties',
            [items[-1].name],
            _("_Track Properties"),
            'document-properties',
            lambda w, n, o, c: o.show_properties_dialog(),
        )
    )

    for item in items:
        providers.register('playlist-context-menu', item)


__create_playlist_context_menu()


class PlaylistPageBase(NotebookPage):
    """
    Base class for playlist pages. Subclasses can indicate that
    they support the following operations:

    save:
        - Define a function called 'on_save'

    save as:
        - Define a function called 'on_saveas'
    """

    menu_provider_name = 'playlist-tab-context-menu'

    def __init__(self, playlist, player):
        NotebookPage.__init__(self)

        self.player = player
        self.playlist = playlist

        self.play_pixbuf = icons.MANAGER.pixbuf_from_icon_name(
            'media-playback-start', size=Gtk.IconSize.MENU
        )
        self.pause_pixbuf = icons.MANAGER.pixbuf_from_icon_name(
            'media-playback-pause', size=Gtk.IconSize.MENU
        )

        event.add_ui_callback(
            self.on_playback_state_change,
            "playback_track_start",
            player,
            destroy_with=self,
        )
        event.add_ui_callback(
            self.on_playback_state_change,
            "playback_track_end",
            player,
            destroy_with=self,
        )
        event.add_ui_callback(
            self.on_playback_state_change,
            "playback_player_pause",
            player,
            destroy_with=self,
        )
        event.add_ui_callback(
            self.on_playback_state_change,
            "playback_player_resume",
            player,
            destroy_with=self,
        )

    def on_playback_state_change(self, typ, player, track):
        """
        Sets the tab icon to reflect the playback status
        """
        if player.queue.current_playlist != self.playlist:
            self.tab.set_icon(None)
        elif typ in ('playback_player_end', 'playback_track_end'):
            self.tab.set_icon(None)
        elif typ in ('playback_track_start', 'playback_player_resume'):
            self.tab.set_icon(self.play_pixbuf)
        elif typ == 'playback_player_pause':
            self.tab.set_icon(self.pause_pixbuf)

    def can_save(self):
        return hasattr(self, 'on_save')

    def can_saveas(self):
        return hasattr(self, 'on_saveas')


class PlaylistPage(PlaylistPageBase):
    """
    Displays a playlist and associated controls.
    """

    def __init__(self, playlist, player):
        """
        :param playlist: The :class:`xl.playlist.Playlist` to display
            in this page.
        :param player: The :class:`xl.player._base.ExailePlayer` that
            this page is associated with
        :param queue:
        """
        PlaylistPageBase.__init__(self, playlist, player)
        NotebookPage.__init__(self)

        self.icon = None

        self.loading = None
        self.loading_timer = None

        uifile = xdg.get_data_path("ui", "playlist.ui")

        self.builder = guiutil.get_builder(uifile)
        playlist_page = self.builder.get_object("playlist_page")

        for child in playlist_page.get_children():
            packing = playlist_page.query_child_packing(child)
            playlist_page.remove(child)
            self.add(child)
            self.set_child_packing(child, *packing)

        self.shuffle_button = self.builder.get_object("shuffle_button")
        self.repeat_button = self.builder.get_object("repeat_button")
        self.dynamic_button = self.builder.get_object("dynamic_button")
        self.search_entry = guiutil.SearchEntry(self.builder.get_object("search_entry"))

        self.builder.connect_signals(self)

        self.playlist_window = self.builder.get_object("playlist_window")
        self.playlist_utilities_bar = self.builder.get_object('playlist_utilities_bar')

        self.view = PlaylistView(playlist, player)
        self.view.set_search_entry(self.search_entry.entry)
        self.view.connect(
            'start-interactive-search', lambda *a: self.search_entry.entry.grab_focus()
        )

        self.playlist_window.add(self.view)

        event.add_ui_callback(
            self.on_mode_changed,
            'playlist_shuffle_mode_changed',
            self.playlist,
            self.shuffle_button,
            destroy_with=self,
        )
        event.add_ui_callback(
            self.on_mode_changed,
            'playlist_repeat_mode_changed',
            self.playlist,
            self.repeat_button,
            destroy_with=self,
        )
        event.add_ui_callback(
            self.on_mode_changed,
            'playlist_dynamic_mode_changed',
            self.playlist,
            self.dynamic_button,
            destroy_with=self,
        )
        event.add_ui_callback(
            self.on_dynamic_playlists_provider_changed,
            'dynamic_playlists_provider_added',
            destroy_with=self,
        )
        event.add_ui_callback(
            self.on_dynamic_playlists_provider_changed,
            'dynamic_playlists_provider_removed',
            destroy_with=self,
        )
        event.add_ui_callback(self.on_option_set, 'gui_option_set', destroy_with=self)

        self.on_mode_changed(
            None, None, self.playlist.shuffle_mode, self.shuffle_button
        )
        self.on_mode_changed(None, None, self.playlist.repeat_mode, self.repeat_button)
        self.on_mode_changed(
            None, None, self.playlist.dynamic_mode, self.dynamic_button
        )
        self.on_dynamic_playlists_provider_changed(None, None, None)
        self.on_option_set(
            'gui_option_set', settings, 'gui/playlist_utilities_bar_visible'
        )
        self.view.connect('button-press-event', self.on_view_button_press_event)
        self.view.model.connect('data-loading', self.on_data_loading)

        if self.view.model.data_loading:
            self.on_data_loading(None, True)

        self.show_all()

    ## NotebookPage API ##

    def focus(self):
        self.view.grab_focus()

    def get_page_name(self):
        return self.playlist.name

    def set_page_name(self, name):
        self.playlist.name = name
        self.name_changed()

    def get_search_entry(self):
        return self.search_entry

    ## End NotebookPage ##

    ## PlaylistPageBase API ##

    # TODO: These two probably shouldn't reach back to main..
    def on_save(self):
        main.exaile().playlists.save_playlist(self.playlist, overwrite=True)

    def on_saveas(self):
        exaile = main.exaile()
        playlists = exaile.playlists
        name = dialogs.ask_for_playlist_name(
            exaile.gui.main.window, playlists, self.playlist.name
        )
        if name is not None:
            self.set_page_name(name)
            playlists.save_playlist(self.playlist)

    ## End PlaylistPageBase API ##

    def on_shuffle_button_press_event(self, widget, event):
        self.__show_toggle_menu(
            Playlist.shuffle_modes,
            Playlist.shuffle_mode_names,
            self.on_shuffle_mode_set,
            'shuffle_mode',
            widget,
            event,
        )

    def on_shuffle_button_popup_menu(self, widget):
        self.__show_toggle_menu(
            Playlist.shuffle_modes,
            Playlist.shuffle_mode_names,
            self.on_shuffle_mode_set,
            'shuffle_mode',
            widget,
            None,
        )
        return True

    def on_repeat_button_press_event(self, widget, event):
        self.__show_toggle_menu(
            Playlist.repeat_modes,
            Playlist.repeat_mode_names,
            self.on_repeat_mode_set,
            'repeat_mode',
            widget,
            event,
        )

    def on_repeat_button_popup_menu(self, widget):
        self.__show_toggle_menu(
            Playlist.repeat_modes,
            Playlist.repeat_mode_names,
            self.on_repeat_mode_set,
            'repeat_mode',
            widget,
            None,
        )
        return True

    def on_dynamic_button_toggled(self, widget):
        if widget.get_active():
            self.playlist.dynamic_mode = self.playlist.dynamic_modes[1]
        else:
            self.playlist.dynamic_mode = self.playlist.dynamic_modes[0]

    def on_search_entry_activate(self, entry):
        filter_string = entry.get_text()
        self.view.filter_tracks(filter_string or None)

    def __show_toggle_menu(self, names, display_names, callback, attr, widget, event):
        """
        Display the menu on the shuffle/repeat toggle buttons

        :param names: The list of names of the menu entries
        :param display_names: The list of names to display on
            each menu entry.
        :param callback: The function to call when a menu item is
            activated. It will be passed the name of the activated item.
        :param attr: The attribute of self.playlist to look at to
            determine the currently-selected item.
        :param widget: The ToggleButton to display the menu on
        :param event: The gtk event that triggered the menu display
        """
        widget.set_active(True)
        menu = Gtk.Menu()
        menu.connect('deactivate', self._mode_menu_set_toggle, widget, attr)
        prev = None
        mode = getattr(self.playlist, attr)
        for name, disp in zip(names, display_names):
            group = None if prev is None else prev.get_group()
            item = Gtk.RadioMenuItem.new_with_mnemonic(group, disp)
            if name == mode:
                item.set_active(True)
            item.connect('activate', callback, name)
            menu.append(item)
            if prev is None:
                menu.append(Gtk.SeparatorMenuItem())
            prev = item
        menu.attach_to_widget(widget)
        menu.show_all()
        if event is not None:
            menu.popup(
                None, None, guiutil.position_menu, widget, event.button, event.time
            )
        else:
            menu.popup(None, None, guiutil.position_menu, widget, 0, 0)
        menu.reposition()

    def _mode_menu_set_toggle(self, menu, button, name):
        mode = getattr(self.playlist, name)
        self.on_mode_changed(None, None, mode, button)

    def on_shuffle_mode_set(self, widget, mode):
        """
        Callback for the Shuffle mode menu
        """
        if widget.get_active():
            self.playlist.shuffle_mode = mode

    def on_repeat_mode_set(self, widget, mode):
        """
        Callback for the Repeat mode menu
        """
        if widget.get_active():
            self.playlist.repeat_mode = mode

    def on_mode_changed(self, evtype, playlist, mode, button):
        button.set_active(mode != 'disabled')

    def on_dynamic_playlists_provider_changed(self, evtype, manager, provider):
        """
        Updates the dynamic button on provider changes
        """
        providers_available = len(providers.get('dynamic_playlists')) > 0
        sensitive = False
        tooltip_text = _('Requires plugins providing dynamic playlists')

        if providers_available:
            sensitive = True
            tooltip_text = _('Dynamically add similar tracks to the playlist')

        self.dynamic_button.set_sensitive(sensitive)
        self.dynamic_button.set_tooltip_text(tooltip_text)

    def on_option_set(self, evtype, settings, option):
        """
        Handles option changes
        """
        if option == 'gui/playlist_utilities_bar_visible':
            visible = settings.get_option(option, True)
            self.playlist_utilities_bar.set_visible(visible)
            self.playlist_utilities_bar.set_sensitive(visible)
            self.playlist_utilities_bar.set_no_show_all(not visible)

    def on_data_loading(self, model, loading):
        '''Called when tracks are being loaded into the model'''
        if loading:
            if self.loading is None and self.loading_timer is None:
                self.view.set_model(None)
                self.loading_timer = GLib.timeout_add(500, self.on_data_loading_timer)
        else:
            if self.loading_timer is not None:
                GLib.source_remove(self.loading_timer)
                self.loading_timer = None

            self.view.set_model(self.view.modelfilter)

            if self.loading is not None:
                guiutil.gtk_widget_replace(self.loading, self.playlist_window)
                self.loading.destroy()
                self.loading = None

    def on_data_loading_timer(self):
        if self.loading_timer is None:
            return

        self.loading_timer = None

        grid = Gtk.Grid()
        sp = Gtk.Spinner()
        sp.start()
        lbl = Gtk.Label.new(_('Loading'))

        grid.attach(sp, 0, 0, 1, 1)
        grid.attach(lbl, 1, 0, 1, 1)

        grid.set_halign(Gtk.Align.CENTER)
        grid.set_valign(Gtk.Align.CENTER)

        self.loading = grid

        guiutil.gtk_widget_replace(self.playlist_window, self.loading)
        self.loading.show_all()

    def on_view_button_press_event(self, view, e):
        """
        Displays the tab context menu upon
        clicks in the contained view
        """
        path = view.get_path_at_pos(int(e.x), int(e.y))
        # We only need the tree path if present
        path = path[0] if path else None

        if (
            not path
            and e.type == Gdk.EventType.BUTTON_PRESS
            and e.triggers_context_menu()
        ):
            self.tab_menu.popup(None, None, None, None, e.button, e.time)


class PlaylistView(AutoScrollTreeView, providers.ProviderHandler):
    __gsignals__ = {}

    def __init__(self, playlist, player):
        AutoScrollTreeView.__init__(self)
        providers.ProviderHandler.__init__(self, 'playlist-columns')

        self.playlist = playlist
        self.player = player

        self._current_vertical_scroll = 0
        self._insert_row = -1
        self._insert_count = 0

        self.menu = PlaylistContextMenu(self)
        self.header_menu = menu.ProviderMenu('playlist-columns-menu', self)
        self.header_menu.set_context_func(self.get_header_context)
        self.header_menu.attach_to_widget(self)

        self.dragging = False
        self.pending_event = None

        self.pending_edit_id = None  # Timeout to ensure a double-click didn't occur
        self.pending_edit_data = None  # (path, col) or None
        self.pending_edit_ok = False  # Set to True by release or timeout

        self._insert_focusing = False

        self._hack_is_osx = sys.platform == 'darwin'
        self._hack_osx_control_mask = False

        # Set to true if you only want things to be copied here, not moved
        self.dragdrop_copyonly = False

        self.set_fixed_height_mode(True)  # MASSIVE speedup - don't disable this!
        self.set_enable_search(True)
        self.selection = self.get_selection()
        self.selection.set_mode(Gtk.SelectionMode.MULTIPLE)

        self._filter_matcher = None

        self._sort_columns = list(common.BASE_SORT_TAGS)  # Column sort order

        self._setup_models()
        self._setup_columns()
        self.columns_changed_id = self.connect(
            "columns-changed", self.on_columns_changed
        )

        self.targets = [
            Gtk.TargetEntry.new("exaile-index-list", Gtk.TargetFlags.SAME_APP, 0),
            Gtk.TargetEntry.new("text/uri-list", 0, 0),
        ]
        self.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            self.targets,
            Gdk.DragAction.COPY | Gdk.DragAction.MOVE,
        )
        self.drag_dest_set(
            Gtk.DestDefaults.ALL,
            self.targets,
            Gdk.DragAction.COPY | Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE,
        )

        event.add_ui_callback(self.on_option_set, "gui_option_set", destroy_with=self)
        event.add_ui_callback(
            self.on_playback_start,
            "playback_track_start",
            self.player,
            destroy_with=self,
        )
        event.add_ui_callback(
            self.on_playback_start,
            "playlist_track_next",
            self.player,
            destroy_with=self,
        )
        self._cursor_changed = self.connect("cursor-changed", self.on_cursor_changed)
        self.connect("row-activated", self.on_row_activated)
        self.connect("key-press-event", self.on_key_press_event)

        self.connect("drag-begin", self.on_drag_begin)
        self.connect("drag-drop", self.on_drag_drop)
        self.connect("drag-data-get", self.on_drag_data_get)
        self.connect("drag-data-received", self.on_drag_data_received)
        self.connect("drag-data-delete", self.on_drag_data_delete)
        self.connect("drag-end", self.on_drag_end)
        self.connect("drag-motion", self.on_drag_motion)
        self.connect("size-allocate", self.on_size_allocate)

    def do_destroy(self):
        # if this isn't disconnected, then the columns are emptied out and
        # the user's settings are overwritten with an empty list
        self.disconnect(self.columns_changed_id)
        AutoScrollTreeView.do_destroy(self)

    def _refilter(self):
        with guiutil.without_model(self):
            self.modelfilter.refilter()

    def filter_tracks(self, filter_string):
        """
        Only show tracks that match the filter. If filter is None, then
        clear any existing filters.

        The filter will search any currently enabled columns AND the
        default columns.
        """

        if filter_string is None:
            self._filter_matcher = None
            self._refilter()
        else:
            # Merge default columns and currently enabled columns
            keyword_tags = set(
                playlist_columns.DEFAULT_COLUMNS
                + [c.name for c in self.get_columns()[1:]]
            )
            self._filter_matcher = trax.TracksMatcher(
                filter_string, case_sensitive=False, keyword_tags=keyword_tags
            )
            logger.debug(
                "Filtering playlist %r by %r.", self.playlist.name, filter_string
            )
            self._refilter()
            logger.debug(
                "Filtering playlist %r by %r completed.",
                self.playlist.name,
                filter_string,
            )

    def get_selection_count(self):
        """
        Returns the number of items currently selected in the
        playlist. Prefer this to len(get_selected_tracks()) et al
        if you will discard the actual track list
        """
        return self.get_selection().count_selected_rows()

    def get_selected_tracks(self):
        """
        Returns a list of :class:`xl.trax.Track`
        which are currently selected in the playlist.
        """
        return [x[1] for x in self.get_selected_items()]

    def get_selected_paths(self):
        """
        Returns a list of pairs of treepaths which are currently
        selected in the playlist.

        The treepaths are returned for the base model, so they are
        indices that can be used with the playlist currently
        associated with this view.
        """
        selection = self.get_selection()
        model, paths = selection.get_selected_rows()

        if isinstance(model, Gtk.TreeModelFilter):
            paths = [model.convert_path_to_child_path(path) for path in paths]

        return paths

    def get_selected_items(self):
        """
        Returns a list of pairs of indices and :class:`xl.trax.Track`
        which are currently selected in the playlist.

        The indices can be used with the playlist currently associated
        with this view.
        """
        selection = self.get_selection()
        model, paths = selection.get_selected_rows()
        try:
            if isinstance(model, Gtk.TreeModelFilter):
                tracks = [
                    (
                        model.convert_path_to_child_path(path)[0],
                        model.get_value(model.get_iter(path), 0),
                    )
                    for path in paths
                ]
            else:
                tracks = [
                    (path[0], model.get_value(model.get_iter(path), 0))
                    for path in paths
                ]
        except TypeError:  # one of the paths was invalid
            return []
        return tracks

    def get_sort_column(self):
        for col in self.get_columns():
            if col.get_sort_indicator():
                return col
        return None

    def get_sort_by(self):
        sortcol = self.get_sort_column()
        if sortcol:
            reverse = sortcol.get_sort_order() == Gtk.SortType.DESCENDING
        else:
            reverse = False
        return (self._sort_columns, reverse)

    def play_track_at(self, position, track):
        """
        When called, this will begin playback of a track at a given
        position in the internal playlist
        """
        self._play_track_at(position, track)

    def _play_track_at(
        self, position, track, on_activated=False, restart_if_playing=False
    ):
        '''Internal API'''
        if not settings.get_option('playlist/enqueue_by_default', False) or (
            self.playlist is self.player.queue and on_activated
        ):
            # If track is already playing, then:
            # a) if restart_if_playing flag is set, restart the track
            # b) otherwise, pause/resume it instead of starting over
            if (
                on_activated
                and not self.player.is_stopped()
                and self.player.queue.current_playlist is self.playlist
                and self.playlist.get_current_position() == position
                and not (restart_if_playing and self.player.is_playing())
            ):
                self.player.toggle_pause()

            elif self.player.queue.is_play_enabled():
                self.playlist.set_current_position(position)
                self.player.queue.set_current_playlist(self.playlist)
                self.player.queue.play(track=track)
        elif self.playlist is not self.player.queue:
            self.player.queue.append(track)

    def _setup_columns(self):
        columns = settings.get_option('gui/columns', playlist_columns.DEFAULT_COLUMNS)
        provider_names = [p.name for p in providers.get('playlist-columns')]
        columns = [name for name in columns if name in provider_names]

        if not columns:
            columns = playlist_columns.DEFAULT_COLUMNS

        self.model._set_columns(columns)

        font, ratio = self._compute_font()

        # create a fixed column for the playlist icon
        self._setup_fixed_column(ratio)

        for column in columns:
            playlist_column = providers.get_provider('playlist-columns', column)(
                self, self.player, font, ratio
            )
            playlist_column.connect('clicked', self.on_column_clicked)

            if isinstance(playlist_column.cellrenderer, Gtk.CellRendererText):
                playlist_column.set_attributes(
                    playlist_column.cellrenderer,
                    sensitive=PlaylistModel.COL_SENSITIVE,
                    weight=PlaylistModel.COL_WEIGHT,
                )
            else:
                playlist_column.set_attributes(
                    playlist_column.cellrenderer, sensitive=PlaylistModel.COL_SENSITIVE
                )

            self.append_column(playlist_column)
            header = playlist_column.get_widget()
            header.show()
            header.get_ancestor(Gtk.Button).connect(
                'button-press-event', self.on_header_button_press, playlist_column
            )

            header.get_ancestor(Gtk.Button).connect(
                'key-press-event', self.on_header_key_press_event, playlist_column
            )

    def _compute_font(self):
        font = settings.get_option('gui/playlist_font', None)
        if font is not None:
            font = Pango.FontDescription(font)

        default_font = Gtk.Widget.get_default_style().font_desc
        if font is None:
            font = default_font

        def_font_sz = default_font.get_size()

        # how much has the font deviated from normal?
        ratio = font.get_size() / def_font_sz

        # small fonts can be problematic..
        # -> TODO: perhaps default widths could be specified
        #          in character widths instead? then we could
        #          calculate it instead of using arbitrary widths
        if ratio < 1:
            ratio = ratio * 1.25

        return font, ratio

    def _setup_fixed_column(self, ratio):
        sz = Gtk.icon_size_lookup(Gtk.IconSize.BUTTON)[1]
        sz = max(int(sz * ratio), 1)

        cell = Gtk.CellRendererPixbuf()
        cell.set_property('xalign', 0.0)

        col = Gtk.TreeViewColumn('')
        col.pack_start(cell, False)
        col.set_max_width(sz)
        col.set_attributes(cell, pixbuf=PlaylistModel.COL_PIXBUF)
        col.set_resizable(False)
        col.set_reorderable(False)
        col.set_clickable(False)
        col.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        self.append_column(col)

    def _on_resize_columns(self):
        for col in self.cols:
            col._setup_sizing()
        pass

    def _refresh_columns(self):
        selection = self.get_selection()
        if not selection:  # The widget has been destroyed
            return

        with self.handler_block(self.columns_changed_id):
            columns = self.get_columns()
            for col in columns:
                self.remove_column(col)
                col.destroyed = True

            self._setup_columns()

        self.queue_draw()

    def _setup_models(self):
        self.model = PlaylistModel(self.playlist, [], self.player, self)
        self.__setup_model_hook = self.model.connect(
            'data-loading', self._on_after_model_loading
        )

        self.modelfilter = self.model.filter_new()
        self.modelfilter.set_visible_func(self._modelfilter_visible_func)
        self.set_model(self.modelfilter)

    def _on_after_model_loading(self, ar1, now_loading):
        """
        This is necessary to ensure that row-insert is connected after the initial loading
        due to threaded PlaylistModel::_load_data_thread and playlists greater than 500 tracks
        """
        if not now_loading:
            self.model.connect('row-inserted', self.on_row_inserted)
            self.model.disconnect(self.__setup_model_hook)

    def _modelfilter_visible_func(self, model, iter, data):
        if self._filter_matcher is not None:
            track = model.get_value(iter, 0)
            return self._filter_matcher.match(trax.SearchResultTrack(track))
        return True

    def on_header_button_press(
        self,
        widget: Gtk.Widget,
        event: Gdk.EventButton,
        header_column: playlist_columns.Column,
    ):
        if event.triggers_context_menu():
            self.set_selected_header(header_column)
            self.header_menu.popup_at_pointer(event)
            return True

    def on_header_key_press_event(
        self,
        widget: Gtk.Widget,
        event: Gdk.EventKey,
        header_column: playlist_columns.Column,
    ):
        if event.keyval == Gdk.KEY_Menu:
            # Open context menu for selecting visible columns
            self.set_selected_header(header_column)
            self.header_menu.popup_at_widget(widget, event)
            return True

    def get_header_context(self, dummy):
        context = {}
        context['trigger_column'] = self.selected_header
        return context

    def set_selected_header(self, header: playlist_columns.Column) -> None:
        self.selected_header = header

    def on_columns_changed(self, widget):
        columns = [c.name for c in self.get_columns()[1:]]
        if columns != settings.get_option('gui/columns', []):
            settings.set_option('gui/columns', columns)

    def on_column_clicked(self, column):
        if self.model.data_loading:
            return
        order = None
        for col in self.get_columns()[1:]:
            if col.name == column.name:
                order = column.get_sort_order()
                if order == Gtk.SortType.ASCENDING:
                    order = Gtk.SortType.DESCENDING
                else:
                    order = Gtk.SortType.ASCENDING
                col.set_sort_indicator(True)
                col.set_sort_order(order)

                self._sort_columns.insert(0, column.name)
            else:
                col.set_sort_indicator(False)
                col.set_sort_order(Gtk.SortType.DESCENDING)
        reverse = order == Gtk.SortType.DESCENDING

        # Uniquify the list of sort columns
        def unique_ordered(seq):
            seen = set()
            return [x for x in seq if x not in seen and not seen.add(x)]

        self._sort_columns = unique_ordered(self._sort_columns)

        # If you don't unselect before sorting, a large number of selected will
        # kill performance
        self.get_selection().unselect_all()
        with self.handler_block(self._cursor_changed):
            self.playlist.sort(self._sort_columns, reverse=reverse)

    def on_option_set(self, typ, obj, data):
        if data == "gui/columns" or data == 'gui/playlist_font':
            self._refresh_columns()

    def on_playback_start(self, type, player, track):
        if (
            player.queue.current_playlist == self.playlist
            and player.current == self.playlist.current
            and settings.get_option('gui/ensure_visible', True)
        ):
            self.scroll_to_current()

    def scroll_to_current(self):
        position = self.playlist.current_position
        if position >= 0:
            model = self.get_model()
            # If it's a filter, then the position isn't actually the path
            if hasattr(model, 'convert_child_path_to_path'):
                path = model.convert_child_path_to_path(Gtk.TreePath((position,)))
            if path:
                self.scroll_to_cell(path)
                self.set_cursor(path)

    def on_cursor_changed(self, widget):
        context = common.LazyDict(self)
        context['selection-empty'] = (
            lambda name, parent: parent.get_selection_count() == 0
        )
        context['selected-items'] = lambda name, parent: parent.get_selected_items()
        context['selected-tracks'] = lambda name, parent: parent.get_selected_tracks()
        context['selection-count'] = lambda name, parent: parent.get_selection_count()
        event.log_event('playlist_cursor_changed', self, context)

    def on_row_activated(self, *args):
        try:
            position, track = self.get_selected_items()[0]
        except IndexError:
            return

        # Call with restart_if_playing flag set to True, so that
        # currently-playing track will be restarted
        self._play_track_at(position, track, True, True)

    def on_row_inserted(self, model, path, iter):
        """
        When something is inserted into the playlist, focus on it. If
        there are multiple things inserted, focus only on the first.
        """
        if not self._insert_focusing:
            self._insert_focusing = True
            # HACK: GI: We get a segfault if we don't do this, because the
            # GtkTreePath gets deleted before the idle function is run.
            path = path.copy()

            def _set_cursor():
                self.set_cursor(path)
                self._insert_focusing = False

            GLib.idle_add(_set_cursor)

    def do_button_press_event(self, e):
        """
        Adds some custom selection work to
        1) unselect all rows if clicking an empty area,
        2) updating the selection upon right click and
        3) popping up the context menu upon right click

        Also sets the internal state for button pressed

        Taken from the following sources:

        * thunar_details_view_button_press_event() of thunar-details-view.c
        * MultiDragTreeView.__button_press/__button.release of quodlibet/qltk/views.py
        """

        # need this to workaround bug in GTK+ on OSX when dragging/dropping
        # -> https://bugzilla.gnome.org/show_bug.cgi?id=722815
        if self._hack_is_osx:
            self._hack_osx_control_mask = (
                True if e.state & Gdk.ModifierType.CONTROL_MASK else False
            )

        selection = self.get_selection()
        pathtuple = self.get_path_at_pos(int(e.x), int(e.y))
        # We only need the tree path if present
        if pathtuple:
            path = pathtuple[0]
            col = pathtuple[1]
        else:
            path = None

        # We unselect all selected items if the user clicks on an empty
        # area of the treeview and no modifier key is active
        if not e.state & Gtk.accelerator_get_default_mod_mask() and not path:
            selection.unselect_all()

        if path and e.type == Gdk.EventType.BUTTON_PRESS:
            # Prevent unselection of all except the clicked item on left
            # clicks, required to preserve the selection for DnD
            if (
                e.button == Gdk.BUTTON_PRIMARY
                and not e.state & Gtk.accelerator_get_default_mod_mask()
                and selection.path_is_selected(path)
            ):
                if selection.count_selected_rows() > 1:
                    selection.set_select_function(lambda *args: False, None)
                    self.pending_event = (path, col)
                elif hasattr(col, 'could_edit') and col.could_edit(self, path, e):
                    # TODO: should cache this?
                    tm = self.get_settings().props.gtk_double_click_time
                    if self.pending_edit_id:
                        GLib.source_remove(self.pending_edit_id)
                    self.pending_edit_data = (path, col)
                    self.pending_edit_ok = False
                    self.pending_edit_id = GLib.timeout_add(
                        tm, self._editing_double_click_ok
                    )

            if e.triggers_context_menu():
                # Select the path on which the user clicked if not selected yet
                if not selection.path_is_selected(path):
                    # We don't unselect all other items if Control is active
                    if not e.state & Gdk.ModifierType.CONTROL_MASK:
                        selection.unselect_all()

                    selection.select_path(path)

                self.menu.popup(None, None, None, None, e.button, e.time)

                return True

        elif e.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
            self._cancel_editing()

        return Gtk.TreeView.do_button_press_event(self, e)

    def do_button_release_event(self, e):
        """
        Unsets the internal state for button press
        """
        self._hack_osx_control_mask = False

        # Restore regular selection behavior in any case
        self.get_selection().set_select_function(lambda *args: True, None)

        if self.pending_event:
            path, col = self.pending_event
            # perform the normal selection that would have happened
            self.set_cursor(path, col, 0)
            self.pending_event = None

        self._maybe_start_editing()

        return Gtk.TreeView.do_button_release_event(self, e)

    def _cancel_editing(self):
        self.pending_edit_ok = False
        self.pending_edit_data = None
        if self.pending_edit_id:
            GLib.source_remove(self.pending_edit_id)
            self.pending_edit_id = None

    def _editing_double_click_ok(self):
        self.pending_edit_id = None
        self._maybe_start_editing()
        return False

    def _maybe_start_editing(self):
        # this has to be called twice for an edit to start
        # -> either by the double click timeout, or button release
        if self.pending_edit_ok:
            path, col = self.pending_edit_data
            col.start_editing(self, path)
            self.pending_edit_ok = False
            self.pending_edit_data = None
        elif self.pending_edit_data:
            self.pending_edit_ok = True

    def on_key_press_event(self, widget, event):
        if event.keyval == Gdk.KEY_Menu:
            self.menu.popup(None, None, None, None, 0, event.time)
            return True

        elif event.keyval == Gdk.KEY_Delete:
            indexes = [x[0] for x in self.get_selected_paths()]
            with guiutil.without_model(self):
                if indexes and indexes == list(
                    range(indexes[0], indexes[0] + len(indexes))
                ):
                    del self.playlist[indexes[0] : indexes[0] + len(indexes)]
                else:
                    for i in indexes[::-1]:
                        del self.playlist[i]

        # TODO: localization?
        # -> Also, would be good to expose these shortcuts somehow to the user...
        #    it seems that Gtk.BindingSet is the way to do it, but there isn't a
        #    mechanism to enumerate it in python
        elif event.keyval == Gdk.KEY_q:
            if self.playlist is not self.player.queue:
                self.player.queue.extend(self.get_selected_tracks())

        elif event.keyval == Gdk.KEY_Up and event.state & Gdk.ModifierType.MOD1_MASK:
            indexes = [x[0] for x in self.get_selected_paths()]
            if len(indexes) == 1 and indexes[0] != 0:
                idx = indexes[0]
                track = self.playlist.pop(idx)
                self.playlist[idx - 1 : idx - 1] = [track]

        elif event.keyval == Gdk.KEY_Down and event.state & Gdk.ModifierType.MOD1_MASK:
            indexes = [x[0] for x in self.get_selected_paths()]
            if len(indexes) == 1 and indexes[0] != len(self.playlist) - 1:
                idx = indexes[0]
                track = self.playlist.pop(idx)
                self.playlist[idx + 1 : idx + 1] = [track]

        elif event.keyval == Gdk.KEY_F2:
            path, column = self.get_cursor()
            if path and isinstance(column, playlist_columns.EditableColumn):
                column.start_editing(self, path)

        # Prevent space-key from triggering 'row-activated' signal,
        # which restarts the currently-playing track. Instead, we call
        # _play_track_at() here with restart_if_playing flag set to
        # False, so that pause will be always toggled instead.
        elif event.keyval == Gdk.KEY_space:
            try:
                position, track = self.get_selected_items()[0]
            except IndexError:
                return True  # Prevent 'row-activated'

            self._play_track_at(position, track, True, False)

            return True  # Prevent 'row-activated'

        # Columns with editable rows will take row-activated and start editing
        # but we want to start playback, so simulate it instead
        elif event.keyval == Gdk.KEY_Return:
            self.on_row_activated()
            return True  # Prevent 'row-activated'

    ### DND handlers ###
    # Source
    def on_drag_begin(self, widget, context):
        """
        Activates the dragging state
        """
        # TODO: set drag icon
        self.dragging = True
        self.pending_event = None
        self._cancel_editing()
        self.get_selection().set_select_function(lambda *args: True, None)

    def on_drag_data_get(self, widget, context, selection, info, etime):
        """
        Stores indices and URIs of the selected items in the drag selection
        """
        target_atom = selection.get_target()
        target = target_atom.name()
        if target == "exaile-index-list":
            positions = self.get_selected_paths()
            if positions:
                s = ",".join(str(i[0]) for i in positions)
                selection.set(target_atom, 8, s.encode('utf-8'))
        elif target == "text/uri-list":
            tracks = self.get_selected_tracks()
            uris = trax.util.get_uris_from_tracks(tracks)
            selection.set_uris(uris)

    def on_drag_data_delete(self, widget, context):
        """
        Stops the default handler from running, all
        processing occurs in the drag-data-received handler
        """
        self.stop_emission_by_name('drag-data-delete')

    def on_drag_end(self, widget, context):
        """
        Deactivates the dragging state
        """
        self.dragging = False

    # Dest
    def on_drag_drop(self, widget, context, x, y, etime):
        """
        Always allows processing of drop operations
        """
        return True

    def on_drag_data_received(self, widget, context, x, y, selection, info, etime):
        """
        Builds a list of tracks either from internal indices or
        external URIs and inserts or appends them to the playlist
        """
        # Stop default handler from running
        self.stop_emission_by_name('drag-data-received')

        # Makes `self.on_row_inserted` to ignore inserted rows
        # see https://github.com/exaile/exaile/issues/487
        self._insert_focusing = True

        drop_info = self.get_dest_row_at_pos(x, y)

        if drop_info:
            path, position = drop_info
            model = self.get_model()
            if isinstance(model, Gtk.TreeModelFilter):
                path = model.convert_path_to_child_path(path)

            insert_position = path[0]
            if position in (
                Gtk.TreeViewDropPosition.AFTER,
                Gtk.TreeViewDropPosition.INTO_OR_AFTER,
            ):
                insert_position += 1
        else:
            insert_position = -1

        # hack: see https://github.com/exaile/exaile/issues/479
        if self._hack_is_osx:
            source_widget = Gtk.drag_get_source_widget(context)
            if not (source_widget is None or source_widget is self) and hasattr(
                source_widget, 'get_selected_tracks'
            ):
                try:
                    tracks = source_widget.get_selected_tracks()
                    if insert_position >= 0:
                        self.playlist[insert_position:insert_position] = tracks
                    else:
                        self.playlist.extend(tracks)
                        insert_position = len(self.playlist) - len(tracks)
                except Exception:
                    pass  # Any exception on hack is ignored
                else:
                    # Select inserted items
                    if tracks:
                        self.selection.unselect_all()
                        self.selection.select_range(
                            self.model.get_path(
                                self.model.iter_nth_child(None, insert_position)
                            ),
                            self.model.get_path(
                                self.model.iter_nth_child(
                                    None, insert_position + len(tracks) - 1
                                )
                            ),
                        )

                    # Restore state to `self.on_row_inserted` do not ignore inserted rows
                    # see https://github.com/exaile/exaile/issues/487
                    self._insert_focusing = False
                    return

        tracks = []
        playlist = []
        positions = []
        pos = 0
        source_playlist_view = None

        target = selection.get_target().name()
        if target == "exaile-index-list":
            selection_data = selection.get_data().decode('utf-8')
            if selection_data == '':  # Ignore drops from empty areas
                # Restore state to `self.on_row_inserted` do not ignore inserted rows
                # see https://github.com/exaile/exaile/issues/487
                self._insert_focusing = False
                return
            positions = [int(pos) for pos in selection_data.split(",")]
            tracks = common.MetadataList()
            source_playlist_view = Gtk.drag_get_source_widget(context)
            playlist = self.playlist

            # Get the playlist of the
            if source_playlist_view is not self:
                playlist = source_playlist_view.playlist

            current_position_index = -1
            for i, pos in enumerate(positions):
                if pos == playlist.current_position:
                    current_position_index = i

                tracks.append(playlist[pos])

            # Insert at specific position if possible
            if insert_position >= 0:
                self.playlist[insert_position:insert_position] = tracks

                if source_playlist_view is self:
                    # Update position for tracks after the insert position
                    for i, position in enumerate(positions[:]):
                        if position >= insert_position:
                            position += len(tracks)
                            positions[i] = position

                # Update playlist current position
                if current_position_index >= 0:
                    self.playlist.current_position = (
                        insert_position + current_position_index
                    )
                    self.model.update_row_params(self.playlist.current_position)
            else:
                # Otherwise just append the tracks
                self.playlist.extend(tracks)
                insert_position = len(self.playlist) - len(tracks)
        elif target == "text/uri-list":
            uris = selection.get_uris()
            tracks = []
            for uri in uris:
                if is_valid_playlist(uri):
                    tracks.extend(import_playlist(uri))
                else:
                    tracks.extend(trax.get_tracks_from_uri(uri))
            sort_by, reverse = self.get_sort_by()
            tracks = trax.sort_tracks(
                sort_by, tracks, reverse=reverse, artist_compilations=True
            )
            if insert_position >= 0:
                self.playlist[insert_position:insert_position] = tracks
            else:
                self.playlist.extend(tracks)
                insert_position = len(self.playlist) - len(tracks)

        # Remove tracks from the source playlist if moved
        if context.get_selected_action() == Gdk.DragAction.MOVE:
            for i in positions[::-1]:
                del playlist[i]

        delete = context.get_selected_action() == Gdk.DragAction.MOVE
        context.finish(True, delete, etime)

        scroll_when_appending_tracks = settings.get_option(
            'gui/scroll_when_appending_tracks', False
        )

        if scroll_when_appending_tracks and tracks:
            self.scroll_to_cell(self.playlist.index(tracks[-1]))
        elif 0 < len(tracks) <= 500:
            # Keep insert position and length of selection for restoring after updating the playlist view
            if (
                0 < pos < insert_position
                and source_playlist_view is self
                and context.get_selected_action() == Gdk.DragAction.MOVE
            ):
                insert_position -= len(tracks)
            self._insert_row = insert_position
            self._insert_count = len(tracks)

        # Restore state to `self.on_row_inserted` do not ignore inserted rows
        # see https://github.com/exaile/exaile/issues/487
        self._insert_focusing = False

    def on_drag_motion(self, widget, context, x, y, etime):
        """
        Makes sure tracks can only be inserted before or after tracks
        and sets the drop action to move or copy depending on target
        and user interaction (e.g. Ctrl key)
        """
        drop_info = self.get_dest_row_at_pos(x, y)

        if not drop_info:
            return False

        path, position = drop_info

        if position == Gtk.TreeViewDropPosition.INTO_OR_BEFORE:
            position = Gtk.TreeViewDropPosition.BEFORE
        elif position == Gtk.TreeViewDropPosition.INTO_OR_AFTER:
            position = Gtk.TreeViewDropPosition.AFTER

        self.set_drag_dest_row(path, position)

        action = Gdk.DragAction.MOVE
        _, _, _, modifier = self.get_window().get_pointer()
        target = self.drag_dest_find_target(context, None).name()

        if (
            target == 'text/uri-list'
            or (self._hack_is_osx and self._hack_osx_control_mask)
            or (not self._hack_is_osx and modifier & Gdk.ModifierType.CONTROL_MASK)
        ):
            action = Gdk.DragAction.COPY

        if self.dragdrop_copyonly and Gtk.drag_get_source_widget(context) != self:
            action = Gdk.DragAction.COPY

        Gdk.drag_status(context, action, etime)

        return True

    def on_size_allocate(self, plView, selection):
        if self._insert_row > -1:
            # correction of selected rows
            self.select_rows(
                self._insert_row, self._insert_row + self._insert_count - 1
            )
            self._insert_row = -1
            self._insert_count = 0

    def show_properties_dialog(self):
        from xlgui import properties

        items = self.get_selected_items()

        # If only one track is selected, we expand `tracks` to include all
        # tracks in the playlist... except for really large playlists, this
        # essentially hangs. Instead, only show all tracks *if* the playlist
        # size is less than 100.
        #
        # A better option would be to lazy load the files, but I'm too lazy
        # to implement that now.. :)

        if len(items) == 1 and len(self.playlist) < 100:
            tracks = self.playlist[:]
            current_position = items[0][0]
        else:
            tracks = [i[1] for i in items]
            current_position = 0

        properties.TrackPropertiesDialog(self.get_toplevel(), tracks, current_position)

    def on_provider_removed(self, provider):
        """
        Called when a column provider is removed
        """
        columns = settings.get_option('gui/columns')

        if provider.name in columns:
            columns.remove(provider.name)
            settings.set_option('gui/columns', columns)

    def select_rows(self, start, stop):
        self.selection.unselect_all()
        if self.model.iter_nth_child(None, start) == None:
            return
        if self.model.iter_nth_child(None, stop) == None:
            return
        self.selection.select_range(
            self.model.get_path(self.model.iter_nth_child(None, start)),
            self.model.get_path(self.model.iter_nth_child(None, stop)),
        )


class PlaylistModel(Gtk.ListStore):
    """
    This ListStore contains all the information needed to render a playlist
    via a PlaylistView. There are five columns:

    * xl.trax.Track
    * dictionary (tag cache)
    * Gdk.Pixbuf (indicates whether track is playing or not)
    * boolean (indicates whether row is sensitive)
    * Pango.Weight (indicates if row is the playing track or not)

    The cache keys correspond to the tags rendered by each column. When a
    track changes, the row's corresponding cache is cleared and the row
    change event is fired.

    The cache keys are populated by the playlist columns. This arrangement
    ensures that we don't have to recreate the playlist model each time the
    columns are changed.
    """

    __gsignals__ = {
        # Called with true indicates starting operation, False ends op
        'data-loading': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_BOOLEAN,))
    }

    COL_TRACK = 0
    COL_CACHE = 1
    COL_PIXBUF = 2
    COL_SENSITIVE = 3
    COL_WEIGHT = 4

    PARAM_COLS = (COL_PIXBUF, COL_SENSITIVE, COL_WEIGHT)

    def __init__(self, playlist, column_names, player, parent):
        # columns: Track, Pixbuf, dict (cache)
        Gtk.ListStore.__init__(
            self, object, object, GdkPixbuf.Pixbuf, bool, Pango.Weight
        )
        self.playlist = playlist
        self.player = player

        self._set_columns(column_names)

        self.data_loading = False
        self.data_load_queue = []

        self._redraw_timer = None
        self._redraw_queue = []

        event.add_ui_callback(
            self.on_tracks_added, "playlist_tracks_added", playlist, destroy_with=parent
        )
        event.add_ui_callback(
            self.on_tracks_removed,
            "playlist_tracks_removed",
            playlist,
            destroy_with=parent,
        )
        event.add_ui_callback(
            self.on_current_position_changed,
            "playlist_current_position_changed",
            playlist,
            destroy_with=parent,
        )
        event.add_ui_callback(
            self.on_spat_position_changed,
            "playlist_spat_position_changed",
            playlist,
            destroy_with=parent,
        )
        event.add_ui_callback(
            self.on_playback_state_change,
            "playback_track_start",
            self.player,
            destroy_with=parent,
        )
        event.add_ui_callback(
            self.on_playback_state_change,
            "playback_track_end",
            self.player,
            destroy_with=parent,
        )
        event.add_ui_callback(
            self.on_playback_state_change,
            "playback_player_pause",
            self.player,
            destroy_with=parent,
        )
        event.add_ui_callback(
            self.on_playback_state_change,
            "playback_player_resume",
            self.player,
            destroy_with=parent,
        )
        event.add_ui_callback(
            self.on_track_tags_changed, "track_tags_changed", destroy_with=parent
        )

        event.add_ui_callback(self.on_option_set, "gui_option_set", destroy_with=parent)

        self._setup_icons()
        self.on_tracks_added(
            None, self.playlist, list(enumerate(self.playlist))
        )  # populate the list

    def _set_columns(self, column_names):
        self.column_names = set(column_names)

    def _setup_icons(self):
        self.play_pixbuf = icons.ExtendedPixbuf(
            icons.MANAGER.pixbuf_from_icon_name('media-playback-start')
        )
        self.pause_pixbuf = icons.ExtendedPixbuf(
            icons.MANAGER.pixbuf_from_icon_name('media-playback-pause')
        )
        self.stop_pixbuf = icons.ExtendedPixbuf(
            icons.MANAGER.pixbuf_from_icon_name('media-playback-stop')
        )
        stop_overlay_pixbuf = self.stop_pixbuf.scale_simple(
            dest_width=self.stop_pixbuf.pixbuf.get_width() // 2,
            dest_height=self.stop_pixbuf.pixbuf.get_height() // 2,
            interp_type=GdkPixbuf.InterpType.BILINEAR,
        )
        stop_overlay_pixbuf = stop_overlay_pixbuf.move(
            offset_x=stop_overlay_pixbuf.pixbuf.get_width(),
            offset_y=stop_overlay_pixbuf.pixbuf.get_height(),
            resize=True,
        )
        self.play_stop_pixbuf = self.play_pixbuf & stop_overlay_pixbuf
        self.pause_stop_pixbuf = self.pause_pixbuf & stop_overlay_pixbuf
        self.clear_pixbuf = self.play_pixbuf.copy()
        self.clear_pixbuf.pixbuf.fill(0x00000000)

        font = settings.get_option('gui/playlist_font', None)
        if font is not None:
            # get default font
            default = float(Gtk.Widget.get_default_style().font_desc.get_size())
            new_font = Pango.FontDescription.from_string(font).get_size()

            # scale pixbuf accordingly
            t = GdkPixbuf.InterpType.BILINEAR
            s = max(int(self.play_pixbuf.get_width() * (new_font / default)), 1)

            self.play_pixbuf = self.play_pixbuf.scale_simple(s, s, t)
            self.pause_pixbuf = self.pause_pixbuf.scale_simple(s, s, t)
            self.stop_pixbuf = self.stop_pixbuf.scale_simple(s, s, t)
            self.play_stop_pixbuf = self.play_stop_pixbuf.scale_simple(s, s, t)
            self.pause_stop_pixbuf = self.pause_stop_pixbuf.scale_simple(s, s, t)
            self.clear_pixbuf = self.clear_pixbuf.scale_simple(s, s, t)

    def _refresh_icons(self):
        self._setup_icons()
        itr = self.get_iter_first()
        position = 0
        while itr:
            self.set(itr, self.PARAM_COLS, self._compute_row_params(position))
            itr = self.iter_next(itr)
            position += 1

    def on_option_set(self, typ, obj, data):
        if data == "gui/playlist_font":
            self._refresh_icons()

    def _compute_row_params(self, rowidx):
        """
        :returns: pixbuf, sensitive, weight
        """

        pixbuf = self.clear_pixbuf.pixbuf
        weight = Pango.Weight.NORMAL
        sensitive = True

        playlist = self.playlist

        spatpos = playlist.spat_position
        spat = spatpos == rowidx

        if spat:
            pixbuf = self.stop_pixbuf.pixbuf

        if playlist is self.player.queue.current_playlist:
            if (
                playlist.current_position == rowidx
                and playlist[rowidx] == self.player.current
            ):
                # this row is the current track, set a special icon
                state = self.player.get_state()

                weight = Pango.Weight.HEAVY

                if state == 'playing':
                    if spat:
                        pixbuf = self.play_stop_pixbuf.pixbuf
                    else:
                        pixbuf = self.play_pixbuf.pixbuf
                elif state == 'paused':
                    if spat:
                        pixbuf = self.pause_stop_pixbuf.pixbuf
                    else:
                        pixbuf = self.pause_pixbuf.pixbuf

            if spatpos == -1 or spatpos > rowidx:
                sensitive = True
            else:
                sensitive = False

        return pixbuf, sensitive, weight

    def update_row_params(self, position):
        itr = self.iter_nth_child(None, position)
        if itr is not None:
            self.set(itr, self.PARAM_COLS, self._compute_row_params(position))

    ### Event callbacks to keep the model in sync with the playlist ###

    def on_tracks_added(self, event_type, playlist, tracks):
        self._load_data(tracks)

    def on_tracks_removed(self, event_type, playlist, tracks):
        for position, track in reversed(tracks):
            if position < 0:
                continue
            self.remove(self.iter_nth_child(None, position))

    def on_current_position_changed(self, event_type, playlist, positions):
        for position in positions:
            if position < 0:
                continue
            self.update_row_params(position)

    def on_spat_position_changed(self, event_type, playlist, positions):
        pos = min(positions)

        if pos < 0:
            pos = 0
            itr = self.get_iter_first()
        else:
            itr = self.iter_nth_child(None, pos)

        while itr:
            self.set(itr, self.PARAM_COLS, self._compute_row_params(pos))
            itr = self.iter_next(itr)
            pos += 1

    def on_playback_state_change(self, event_type, player_obj, track):
        position = self.playlist.current_position
        if position < 0 or position >= len(self):
            return
        self.update_row_params(position)

    def on_track_tags_changed(self, type, track, tags):
        if (
            not track
            or not settings.get_option('gui/sync_on_tag_change', True)
            or not (tags & self.column_names)
        ):
            return

        if self._redraw_timer:
            GLib.source_remove(self._redraw_timer)
        self._redraw_queue.append(track)
        self._redraw_timer = GLib.timeout_add(100, self._on_track_tags_changed)

    def _on_track_tags_changed(self):
        self._redraw_timer = None
        redraw_queue = set(self._redraw_queue)
        self._redraw_queue = []

        COL_TRACK = self.COL_TRACK
        COL_CACHE = self.COL_CACHE

        for row in self:
            if row[COL_TRACK] in redraw_queue:
                row[COL_CACHE].clear()
                self.row_changed(row.path, row.iter)

    #
    # Loading data into the playlist:
    #
    # - Profiler reveals that most of the playlist loading time is spent in two
    #   places
    #   - Formatting values
    #   - Converting them to GValue objects for insert into the treeview
    #
    # Now, the program is annoyingly blocked when this happens, so we need to
    # process new tracks on a different thread, and show some kind of loading
    # indicator instead. That's what these four functions help us do.
    #

    def _load_data(self, tracks):
        # Don't allow race condition between adds.. there's probably a race
        # condition for removal
        if self.data_loading:
            self.data_load_queue.extend(tracks)
            return

        self.data_loading = True
        self.emit('data-loading', True)

        if len(tracks) > 500:
            self._load_data_thread(tracks)
        else:
            render_data = self._load_data_fn(tracks)
            self._load_data_done(render_data)

    @common.threaded
    def _load_data_thread(self, tracks):
        render_data = self._load_data_fn(tracks)
        GLib.idle_add(self._load_data_done, render_data)

    def _load_data_fn(self, tracks):
        indices = (0, 1, 2, 3, 4)
        return [
            (position, indices, (track, {}) + self._compute_row_params(position))
            for position, track in tracks
        ]

    def _load_data_done(self, render_data):
        for args in render_data:
            self.insert_with_valuesv(*args)

        self.data_loading = False
        self.emit('data-loading', False)

        if self.data_load_queue:
            tracks = self.data_load_queue
            self.data_load_queue = []

            self._load_data(tracks)



"""
This module is a parser for audio CDs based on musicbrainzngs and discid.

It uses the online database provided by musicbrainzngs.

musicbrainz documentation:
* Wiki: https://wiki.musicbrainz.org/
* Web API: https://musicbrainz.org/doc/Development/XML%20Web%20Service/Version%202
* Web API format: https://github.com/metabrainz/mmd-schema/blob/master/schema/musicbrainz_mmd-2.0.rng
* Database schema: https://wiki.musicbrainz.org/MusicBrainz_Database/Schema

musicbrainzngs documentation:
* https://python-musicbrainzngs.readthedocs.io/

Note: All the "priority" values are arbitrary/random choices and may need improvement

TODO: Room for improvement:
* fetch composers (especially interesting for classical music)
* test text-representation for non-latin languages
* get more data with secondary query, e.g. genre ("medium_tags")
    see https://wiki.musicbrainz.org/Genre
"""

from __future__ import division

import logging

import musicbrainzngs

from xl import main

from xl.covers import MANAGER as CoverManager

logger = logging.getLogger(__name__)


def fetch_with_disc_id(disc_id):
    """
    I/O operation to fetch data from musicbrainz over the internet.
    This must happen async because it may take quite some time.

    @param disc_id: A musicbrainz disc ID
    @return: Metadata for the parse() function
    """
    musicbrainzngs.set_useragent(
        'Exaile_Cd_Import', main.__version__, 'https://exaile.org/'
    )
    logger.debug('Querying Musicbrainz web servers with disc id %s', disc_id)
    try:
        musicbrainz_data = musicbrainzngs.get_releases_by_discid(
            disc_id.id,
            toc=disc_id.toc_string,
            includes=['artists', 'recordings', 'isrcs', 'artist-credits'],
        )
        logger.debug('Received data from musicbrainzngs')
    except musicbrainzngs.WebServiceError:
        # This is expected to fail if user is offline or behind an
        # aggressive firewall.
        logger.info('Failed to fetch data from musicbrainz database.', exc_info=True)
        musicbrainz_data = None
    return musicbrainz_data


def parse(musicbrainz_data, disc_id, tracks):
    """
    Parses the given musicbrainz data into existing tracks.
    This function must be run on the main thread because it modifies
    tracks, which may be in use by the main thread!

    @param musicbrainz_data: The data from fetch_with_disc_id()
    @param disc_id: The disc ID from read_disc_id()
    @param device: Name of the CD device
    @return: An array of xl.trax.Track with more information
    """
    # TODO: This parser is very slow, especially before creating tracks.
    # Put some of the parser (which does not modify tracks) into separate thread
    if musicbrainz_data is None:
        return None
    disc_tracks = None
    try:
        if musicbrainz_data.get('disc') is not None:  # preferred: good quality
            disc_tracks = __parse_musicbrainz_disc_data(
                musicbrainz_data['disc'], disc_id, tracks
            )
            if disc_tracks is not None:
                logger.debug('Parsed disc data: %s', disc_tracks)
        if disc_tracks is None:
            if (
                musicbrainz_data.get('cdstub') is not None
            ):  # bad quality, use as fallback
                disc_tracks = __parse_musicbrainz_cdstub_data(
                    musicbrainz_data['cdstub'], tracks
                )
                if disc_tracks is not None:
                    logger.debug('Parsed cdstub data: %s', disc_tracks)
        if disc_tracks is None:
            logger.info('Musicbrainz returned no useful data: %s', musicbrainz_data)
    except Exception:
        logger.warn('Failed to parse data from musicbrainz database.', exc_info=True)
    return disc_tracks


def __parse_musicbrainz_cdstub_data(cdstub_data, tracks):
    """
    Parses cdstub data into xl.trax.Track
    See "def_cdstub"
    Unused: disambiguation, def_cdstub-attribute_extension, def_cdstub-element_extension
    """
    track_count = cdstub_data.get('track-count')
    if track_count is not None:
        if len(tracks) is not track_count:
            logger.debug(
                'Track count mismatch: real disc has %i, musicbrainz says %i. '
                'Ignoring all data.',
                len(tracks),
                track_count,
            )
            return None

    album_tags = dict()

    if cdstub_data.get('id') is not None:
        album_tags['__musicbrainz_cdstub_id'] = cdstub_data['id']

    album_title = cdstub_data['title']
    album_tags['album'] = album_title

    artist = cdstub_data.get('artist')
    if artist is not None:
        album_tags['albumartist'] = artist
        album_tags['artist'] = artist

    barcode = cdstub_data.get('barcode')
    if barcode is not None:
        album_tags['barcode'] = barcode

    track_list = cdstub_data.get('track-list')
    if track_list is not None:
        for i in range(0, len(tracks)):
            # See "def_nonmb-track"
            # unused: length
            track = tracks[i]
            track_tags = album_tags.copy()
            track_tags['title'] = track_list[i]['title']

            track_artist = track_list[i].get('artist')
            if track_artist is not None:
                track_tags['artist'] = track_artist

            track.set_tags(**track_tags)
    return tracks, album_title


def __parse_musicbrainz_disc_data(disc_data, disc_id, tracks):
    # get list of potential candidates, throw away the rest
    release_list = disc_data['release-list']
    suitable_releases = []
    for single_release in release_list:
        release = __check_single_release(single_release, disc_id)
        if release is not None:
            suitable_releases.append(release)

    if len(suitable_releases) < 1:
        return None

    release, medium = __choose_release_and_medium(suitable_releases)
    return __parse_medium_from_disc_data(release, medium, tracks)


def __choose_release_and_medium(suitable_releases):
    """
    find the most suitable release
    """
    chosen_prio = -10  # this is a threshold
    chosen_release = None
    chosen_release_media = None
    for prio, release_media, release in suitable_releases:
        if prio > chosen_prio:
            chosen_prio = prio
            chosen_release_media = release_media
            chosen_release = release

    # find the most suitable medium
    chosen_prio = -50  # this is a threshold
    chosen_medium = None
    for prio, medium in chosen_release_media:
        if prio > chosen_prio:
            chosen_prio = prio
            chosen_medium = medium

    return (chosen_release, chosen_medium)


def __check_single_release(single_release, disc_id):
    """
    Documentation: https://wiki.musicbrainz.org/Recording
    """

    priority = 0
    # See https://wiki.musicbrainz.org/Barcode
    mb_barcode = single_release.get('barcode')
    if mb_barcode is not None and disc_id.mcn is not None:
        mb_barcode = mb_barcode.lstrip('0')
        disc_barcode = disc_id.mcn.lstrip('0')

        if len(disc_barcode) > 5:  # empty string indicates missing barcode
            if mb_barcode == disc_barcode:
                logger.debug('Found exact barcode match: %s', disc_barcode)
                priority = 100
            else:  # could be a different barcode format or a weak match
                logger.debug(
                    'Barcode mismatch: real disc has %s, musicbrainz says %s.',
                    disc_barcode,
                    mb_barcode,
                )
                priority = -19

    medium_list = single_release.get('medium-list')
    if medium_list is None:
        logger.debug('Release contains no `medium-list`. This sounds fishy, ignoring.')
        return None

    suitable_media = []
    for single_medium in medium_list:
        medium_priority = __check_single_medium(single_medium, disc_id)
        if medium_priority is not None:
            suitable_media.append((medium_priority, single_medium))
            priority = priority + medium_priority / 20

    if len(suitable_media) == 0:
        return None
    else:
        return (priority, suitable_media, single_release)


def __check_single_medium(single_medium, disc_id):
    """
    Documentation: https://wiki.musicbrainz.org/Medium
    """
    priority = 0
    mb_format = single_medium.get('format')
    if mb_format is not None:
        if not mb_format == 'CD':
            logger.debug(
                'Medium format mismatch: real disc is CD, musicbrainz says %s. Ignoring.',
                mb_format,
            )
            return None
        priority = priority + 7

    track_count = single_medium.get('track-count')
    if track_count is not None:
        if not track_count == len(disc_id.tracks):
            logger.debug(
                'Track count mismatch: real disc has %i, musicbrainz says %i. Ignoring.',
                len(disc_id.tracks),
                track_count,
            )
            return None
        priority = priority + 2

    track_list = single_medium.get('track-list')
    if track_list is None:
        logger.debug('Medium contains no `track-list`. This sounds fishy, ignoring.')
    if not len(track_list) == len(disc_id.tracks):
        logger.debug(
            'Track number mismatch: real disc has %i, musicbrainz says %i. Ignoring.',
            len(disc_id.tracks),
            track_list,
        )
        return None

    # disc-list does not seem to contain important information.

    priority = 0
    for i in range(0, len(track_list)):
        single_mb_track = track_list[i]
        single_disc_track = disc_id.tracks[i]
        track_prio = __check_single_track(single_mb_track, single_disc_track)
        if not track_prio:
            logger.debug('Track mismatch, ignoring this disk.')
            return None
        priority = priority + track_prio
    return priority


def __check_single_track(single_mb_track, single_disc_track):
    """
    See "def_track-data"
    """
    priority = 0

    # TODO: position or number?
    position = single_mb_track.get('position')
    if position is not None:
        mb_position = int(position)
        if mb_position == single_disc_track.number:
            priority = priority + 9
        else:
            logger.info(
                'Track position mismatch: real disc has %i, musicbrainz says %i.',
                single_disc_track.number,
                mb_position,
            )
            return None

    if single_mb_track.get('number') is not None:
        mb_number = int(single_mb_track['number'])
        if mb_number == single_disc_track.number:
            priority = priority + 9
        else:
            logger.info(
                'Track number mismatch: real disc has %i, musicbrainz says %i.',
                single_disc_track.number,
                mb_number,
            )
            return None

    # compare length
    if single_mb_track.get('length') is not None:
        mb_length_ms = int(single_mb_track['length'])
        real_length_ms = single_disc_track.sectors * 1000 / 75
        len_diff = abs(real_length_ms - mb_length_ms)
        if len_diff > 10000:  # up to 5 seconds fade in / fade out time at start and end
            logger.debug(
                'Massive track length mismatch: real disc has %ims, musicbrainz says %ims.',
                real_length_ms,
                mb_length_ms,
            )
            return None
        elif len_diff < 20:  # sector length = 1000ms/75 = (13+1/3)ms
            priority = priority + 50  # priority bonus for "exact" length match
        else:
            priority = priority - len_diff / 100
            logger.debug(
                'Slight track length mismatch: real disc has %i, musicbrainz says %i.',
                real_length_ms,
                mb_length_ms,
            )

    if single_mb_track.get('recording') is not None:
        resource_priority = __check_single_track_as_recording(
            single_mb_track.get('recording'), single_disc_track
        )
        if resource_priority is None:
            return None
        else:
            priority = priority + resource_priority

    return priority


def __check_single_track_as_recording(mb_recording, disc_track):
    """
    See "def_recording-element"
    """

    priority = 0
    if hasattr(disc_track, 'isrc') and disc_track.isrc is not None:
        mb_isrc_list = mb_recording.get('isrc-list')
        if mb_isrc_list is not None:
            if disc_track.isrc in mb_isrc_list:
                priority = +1000
            else:
                priority = -100
    return priority


def __parse_medium_from_disc_data(release, medium, tracks):
    medium_tags = dict()

    disc_number = __get_disc_number(medium, release)
    if disc_number is not None:
        medium_tags['discnumber'] = disc_number

    if release.get('id') is not None:
        medium_tags['__release_id'] = release['id']

    if release.get('date') is not None:
        medium_tags['date'] = release['date']

    # TODO difference of medium.title vs. release.title?
    album_title = medium.get('title')
    if album_title is not None:
        medium_tags['album'] = album_title
    else:
        album_title = release.get('title')
        if album_title is not None:
            medium_tags['album'] = album_title

    artist = release.get('artist-credit-phrase')
    if artist is not None:
        medium_tags['albumartist'] = artist
        medium_tags['artist'] = artist

    if release.get('barcode') is not None:
        medium_tags['barcode'] = release.get('barcode')

    mb_tracks = medium['track-list']
    for track_index in range(0, len(mb_tracks)):
        __parse_track_from_disc_data(
            medium_tags.copy(), tracks[track_index], mb_tracks[track_index]
        )

    # needs the 'musicbrainz_albumid' tag to be set, thus this happens after writing track tags
    __get_cover_image(release, tracks)

    return (tracks, album_title)


def __get_cover_image(release, tracks):
    """
    Fetch a cover using musicbrainzcover's cover provider.
    Requires the 'musicbrainz_albumid' tag to be set on the tracks.

    Documentation:
    * https://musicbrainz.org/doc/Cover_Art_Archive/API
    """
    if release.get('id') is None:
        return None

    try:
        import plugins.musicbrainzcovers as musicbrainzcovers
    except ImportError:
        logger.warn('Cannot load musicbrainzcovers, will not fetch covers.')
        return None

    cover_searcher = musicbrainzcovers.MusicBrainzCoverSearch(main.exaile())
    cover_data = None
    for resolution in ['500', '1200', '250']:
        try:
            db_string = '%s:%s' % (release['id'], resolution)
            cover_data = cover_searcher.get_cover_data(db_string)
            break
        except Exception:
            logger.debug(
                'Cannot fetch cover for %s in resolution %s', release['id'], resolution
            )

    if cover_data is not None:
        for track in tracks:
            # TODO: This is a hack: set_cover() expects a db_string to start with this, but why?
            # We could use CoverManager.get_cover(track) but this does not work because
            # CoverManager.get_db_string(track) fails to get the right string.
            db_string = 'musicbrainz: '
            CoverManager.set_cover(track, db_string, cover_data)


def __parse_track_from_disc_data(track_tags, xl_track, mb_track):
    """
    See "def_track-data" in
    https://github.com/metabrainz/mmd-schema/blob/master/schema/musicbrainz_mmd-2.0.rng
    unused and populated from discid: position, number, length
    """
    if mb_track.get('id') is not None:
        track_tags['__musicbrainz_track_id'] = mb_track['id']

    if mb_track.get('title') is not None:
        track_tags['title'] = mb_track['title']

    track_artist = mb_track.get('artist-credit-phrase')
    if track_artist is not None:
        track_tags['artist'] = track_artist

    if mb_track.get('recording') is not None:
        mb_recording = mb_track['recording']
        # see "def_recording-element"
        # unused fields: length, annotation, disambiguation, ...

        recording_title = mb_recording.get('title')
        if recording_title is not None and track_tags.get('title') is None:
            track_tags['title'] = recording_title

    xl_track.set_tags(**track_tags)


def __get_disc_number(medium, release):
    """
    Helper function to extract disc number string, e.g. 1/2 for the first of 2 CDs
    """
    # We may also be able to use the 'disc-count' field of "medium"
    medium_position = medium.get('position')
    medium_count = release.get('medium-count')
    if medium_count is None:
        medium_count = len(release['medium-list'])
    if medium_position is not None:
        if medium_count is not None:
            return '{0}/{1}'.format(medium_position, medium_count)
        else:
            return medium_position
    elif medium_count is not None and medium_count == 1:
        return '1/1'
    return None



# Copyright (C) 2012  Mathias Brodala <info@noctus.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import urllib.error
import urllib.request

from xl import common, covers, main, providers

import musicbrainzngs

logger = logging.getLogger(__name__)


def enable(exaile):
    """
    Enables the plugin
    """
    providers.register('covers', MusicBrainzCoverSearch(exaile))


def disable(exaile):
    """
    Disables the plugin
    """
    providers.unregister('covers', providers.get_provider('covers', 'musicbrainz'))


class MusicBrainzCoverSearch(covers.CoverSearchMethod):
    """
    Searches MusicBrainz for an album cover
    """

    name = 'musicbrainz'
    title = 'MusicBrainz'
    __caa_url = 'https://coverartarchive.org/release/{mbid}/front-{size}'

    def __init__(self, exaile):
        self.user_agent = exaile.get_user_agent_string('musicbrainzcovers')

    def find_covers(self, track, limit=-1):
        """
        Performs the search
        """
        try:
            artist = track.get_tag_raw('artist')[0]
            album = track.get_tag_raw('album')[0]
        except (AttributeError, TypeError):
            return []

        musicbrainzngs.set_useragent(
            'Exaile_MusicBrainz_Covers', main.__version__, 'https://exaile.org/'
        )
        result = musicbrainzngs.search_releases(
            release=album,
            artistname=artist,
            format='CD',
            limit=3,  # Unlimited search is slow
        )

        if result['release-list']:
            mbids = [a['id'] for a in result['release-list']]

            # Check the actual availability of the covers
            for mbid in mbids[:]:
                try:
                    url = self.__caa_url.format(mbid=mbid, size=250)

                    headers = {'User-Agent': self.user_agent}
                    req = urllib.request.Request(url, None, headers)
                    response = urllib.request.urlopen(req)
                except urllib.error.HTTPError:
                    mbids.remove(mbid)
                else:
                    response.close()

            # For now, limit to small sizes
            mbids = [mbid + ':250' for mbid in mbids]

            return mbids

        return []

    def get_cover_data(self, db_string):
        """
        Get the image data
        """
        data = None
        mbid, size = db_string.split(':')
        url = self.__caa_url.format(mbid=mbid, size=size)

        try:
            logger.debug('Fetching cover from {url}'.format(url=url))
            data = common.get_url_contents(url, self.user_agent)
        except urllib.error.HTTPError:
            pass

        return data



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
Provides methods for convenient icons and image handling
"""

import glob
import logging
import os
from typing import Optional, Union

from gi.repository import GdkPixbuf, GLib, Gtk

from xl import common, covers, settings
from xlgui.guiutil import pixbuf_from_data

logger = logging.getLogger(__name__)


class ExtendedPixbuf:
    """
    A :class:`GdkPixbuf.Pixbuf` wrapper class allowing for
    interaction using standard operators

    Thus you can do the following:

    * ``pixbuf1 + pixbuf2`` (horizontally appends ``pixbuf2`` to ``pixbuf1``)
    * ``pixbuf * 5`` (multiplies the content of ``pixbuf``)
    * ``pixbuf1 & pixbuf2`` (simple composition of ``pixbuf2`` on ``pixbuf1``, the desired alpha value has to be included in the pixbufs themselves)
    * ``pixbuf1 < pixbuf2``, ``pixbuf1 > pixbuf2`` (compares the pixbuf dimensions)
    * ``pixbuf1 == pixbuf2`` (compares the pixel data, use the *is* operator to check for identity)

    Even more is possible with the provided verbose methods
    """

    def __init__(self, pixbuf):
        self.pixbuf = GdkPixbuf.Pixbuf.new(
            pixbuf.get_colorspace(),
            pixbuf.get_has_alpha(),
            pixbuf.get_bits_per_sample(),
            pixbuf.get_width(),
            pixbuf.get_height(),
        )
        pixbuf.copy_area(
            src_x=0,
            src_y=0,
            width=self.pixbuf.get_width(),
            height=self.pixbuf.get_height(),
            dest_pixbuf=self.pixbuf,
            dest_x=0,
            dest_y=0,
        )

    def get_width(self):
        return self.pixbuf.get_width()

    def get_height(self):
        return self.pixbuf.get_height()

    def __add__(self, other):
        """
        Horizontally appends a pixbuf to the current

        :param other: the pixbuf to append
        :type other: :class:`GdkPixbuf.Pixbuf`
        :returns: a new pixbuf
        :rtype: :class:`ExtendedPixbuf`
        """
        return self.add_horizontal(other)

    def __mul__(self, multiplier):
        """
        Horizontally multiplies the current pixbuf content

        :param multiplier: How often the pixbuf
            shall be multiplied
        :type multiplier: int
        :returns: a new pixbuf
        :rtype: :class:`ExtendedPixbuf`
        """
        return self.multiply_horizontal(multiplier)

    def __and__(self, other):
        """
        Composites a pixbuf on the current
        pixbuf at the location (0, 0)

        :param other: the pixbuf to composite
        :type other: :class:`GdkPixbuf.Pixbuf`
        :returns: a new pixbuf
        :rtype: :class:`ExtendedPixbuf`
        """
        return self.composite_simple(other)

    def __lt__(self, other):
        """
        True if the size (width * height) of the current
        pixbuf is lower than the size of the other pixbuf
        """
        if isinstance(other, ExtendedPixbuf):
            other = other.pixbuf
        self_size = self.pixbuf.get_width() * self.pixbuf.get_height()
        other_size = other.get_width() * other.get_height()

        return self_size < other_size

    def __le__(self, other):
        """
        True if the size (width * height) of the current
        pixbuf is lower than or equal to the size of the
        other pixbuf
        """
        if isinstance(other, ExtendedPixbuf):
            other = other.pixbuf
        self_size = self.pixbuf.get_width() * self.pixbuf.get_height()
        other_size = other.get_width() * other.get_height()

        return self_size <= other_size

    def __gt__(self, other):
        """
        True if the size (width * height) of the current
        pixbuf is higher than the size of the other pixbuf
        """
        if isinstance(other, ExtendedPixbuf):
            other = other.pixbuf
        self_size = self.pixbuf.get_width() * self.pixbuf.get_height()
        other_size = other.get_width() * other.get_height()

        return self_size > other_size

    def __ge__(self, other):
        """
        True if the size (width * height) of the current
        pixbuf is higher than or equal to the size of the
        other pixbuf
        """
        if isinstance(other, ExtendedPixbuf):
            other = other.pixbuf
        self_size = self.pixbuf.get_width() * self.pixbuf.get_height()
        other_size = other.get_width() * other.get_height()

        return self_size >= other_size

    def __eq__(self, other):
        """
        True if the pixels of the current pixbuf are
        the same as the pixels from the other pixbuf
        """
        if isinstance(other, ExtendedPixbuf):
            other = other.pixbuf
        return self.pixbuf.get_pixels() == other.get_pixels()

    def __ne__(self, other):
        """
        True if the pixels of the current pixbuf are
        not the same as the pixels from the other pixbuf
        """
        return not self.__eq__(other)

    def add_horizontal(self, other, spacing=0):
        """
        Horizontally appends a pixbuf to the current

        :param other: the pixbuf to append
        :type other: :class:`GdkPixbuf.Pixbuf`
        :param spacing: amount of pixels between the pixbufs
        :type spacing: int
        :returns: a new pixbuf
        :rtype: :class:`ExtendedPixbuf`
        """
        if isinstance(other, ExtendedPixbuf):
            other = other.pixbuf

        height = max(self.pixbuf.get_height(), other.get_height())
        spacing = max(0, spacing)

        new_pixbuf = GdkPixbuf.Pixbuf.new(
            self.pixbuf.get_colorspace(),
            self.pixbuf.get_has_alpha(),
            self.pixbuf.get_bits_per_sample(),
            self.pixbuf.get_width() + spacing + other.get_width(),
            height,
        )
        new_pixbuf.fill(0xFFFFFF00)

        self.pixbuf.copy_area(
            src_x=0,
            src_y=0,
            width=self.pixbuf.get_width(),
            height=self.pixbuf.get_height(),
            dest_pixbuf=new_pixbuf,
            dest_x=0,
            dest_y=0,
        )
        other.copy_area(
            src_x=0,
            src_y=0,
            width=other.get_width(),
            height=other.get_height(),
            dest_pixbuf=new_pixbuf,
            dest_x=self.pixbuf.get_width() + spacing,
            dest_y=0,
        )

        return ExtendedPixbuf(new_pixbuf)

    def add_vertical(self, other, spacing=0):
        """
        Vertically appends a pixbuf to the current

        :param other: the pixbuf to append
        :type other: :class:`GdkPixbuf.Pixbuf`
        :param spacing: amount of pixels between the pixbufs
        :type spacing: int
        :returns: a new pixbuf
        :rtype: :class:`ExtendedPixbuf`
        """
        if isinstance(other, ExtendedPixbuf):
            other = other.pixbuf

        width = max(self.pixbuf.get_width(), other.get_width())
        spacing = max(0, spacing)

        new_pixbuf = GdkPixbuf.Pixbuf.new(
            self.pixbuf.get_colorspace(),
            self.pixbuf.get_has_alpha(),
            self.pixbuf.get_bits_per_sample(),
            width,
            self.pixbuf.get_height() + spacing + other.get_height(),
        )
        new_pixbuf.fill(0xFFFFFF00)

        self.pixbuf.copy_area(
            src_x=0,
            src_y=0,
            width=self.pixbuf.get_width(),
            height=self.pixbuf.get_height(),
            dest_pixbuf=new_pixbuf,
            dest_x=0,
            dest_y=0,
        )
        other.copy_area(
            src_x=0,
            src_y=0,
            width=other.get_width(),
            height=other.get_height(),
            dest_pixbuf=new_pixbuf,
            dest_x=0,
            dest_y=self.pixbuf.get_height() + spacing,
        )

        return ExtendedPixbuf(new_pixbuf)

    def multiply_horizontal(self, multiplier, spacing=0):
        """
        Horizontally multiplies the current pixbuf content

        :param multiplier: How often the pixbuf
            shall be multiplied
        :type multiplier: int
        :param spacing: amount of pixels between the pixbufs
        :type spacing: int
        :returns: a new pixbuf
        :rtype: :class:`ExtendedPixbuf`
        """
        spacing = max(0, spacing)
        new_pixbuf = GdkPixbuf.Pixbuf.new(
            self.pixbuf.get_colorspace(),
            self.pixbuf.get_has_alpha(),
            self.pixbuf.get_bits_per_sample(),
            self.pixbuf.get_width() * multiplier + spacing * multiplier,
            self.pixbuf.get_height(),
        )
        new_pixbuf.fill(0xFFFFFF00)

        for n in range(0, multiplier):
            self.pixbuf.copy_area(
                src_x=0,
                src_y=0,
                width=self.pixbuf.get_width(),
                height=self.pixbuf.get_height(),
                dest_pixbuf=new_pixbuf,
                dest_x=n * self.pixbuf.get_width() + n * spacing,
                dest_y=0,
            )

        return ExtendedPixbuf(new_pixbuf)

    def multiply_vertical(self, multiplier, spacing=0):
        """
        Vertically multiplies the current pixbuf content

        :param multiplier: How often the pixbuf
            shall be multiplied
        :type multiplier: int
        :param spacing: amount of pixels between the pixbufs
        :type spacing: int
        :returns: a new pixbuf
        :rtype: :class:`ExtendedPixbuf`
        """
        spacing = max(0, spacing)
        new_pixbuf = GdkPixbuf.Pixbuf.new(
            self.pixbuf.get_colorspace(),
            self.pixbuf.get_has_alpha(),
            self.pixbuf.get_bits_per_sample(),
            self.pixbuf.get_width(),
            self.pixbuf.get_height() * multiplier + spacing * multiplier,
        )
        new_pixbuf.fill(0xFFFFFF00)

        for n in range(0, multiplier):
            self.pixbuf.copy_area(
                src_x=0,
                src_y=0,
                width=self.pixbuf.get_width(),
                height=self.pixbuf.get_height(),
                dest_pixbuf=new_pixbuf,
                dest_x=0,
                dest_y=n * self.pixbuf.get_height() + n * spacing,
            )

        return ExtendedPixbuf(new_pixbuf)

    def composite_simple(self, other):
        """
        Composites a pixbuf on the current
        pixbuf at the location (0, 0)

        :param other: the pixbuf to composite
        :type other: :class:`GdkPixbuf.Pixbuf`
        :returns: a new pixbuf
        :rtype: :class:`ExtendedPixbuf`
        """
        if isinstance(other, ExtendedPixbuf):
            other = other.pixbuf

        width = max(self.pixbuf.get_width(), other.get_width())
        height = max(self.pixbuf.get_height(), other.get_height())

        new_pixbuf = GdkPixbuf.Pixbuf.new(
            self.pixbuf.get_colorspace(),
            self.pixbuf.get_has_alpha(),
            self.pixbuf.get_bits_per_sample(),
            width,
            height,
        )
        new_pixbuf.fill(0xFFFFFF00)

        for pixbuf in (self.pixbuf, other):
            pixbuf.composite(
                dest=new_pixbuf,
                dest_x=0,
                dest_y=0,
                dest_width=pixbuf.get_width(),
                dest_height=pixbuf.get_height(),
                offset_x=0,
                offset_y=0,
                scale_x=1,
                scale_y=1,
                interp_type=GdkPixbuf.InterpType.BILINEAR,
                # Alpha needs to be embedded in the pixbufs
                overall_alpha=255,
            )

        return ExtendedPixbuf(new_pixbuf)

    def move(self, offset_x, offset_y, resize=False):
        """
        Moves the content of the current pixbuf within
        its boundaries (clips overlapping data) and
        optionally resizes the pixbuf to contain the
        movement

        :param offset_x: the amount of pixels to move
            in horizontal direction
        :type offset_x: int
        :param offset_y: the amount of pixels to move
            in vertical direction
        :type offset_y: int
        :param resize: whether to resize the pixbuf
            on movement
        :type resize: bool
        :returns: a new pixbuf
        :rtype: :class:`ExtendedPixbuf`
        """
        width = self.pixbuf.get_width()
        height = self.pixbuf.get_height()

        if resize:
            width += offset_x
            height += offset_y

        new_pixbuf = GdkPixbuf.Pixbuf.new(
            self.pixbuf.get_colorspace(),
            self.pixbuf.get_has_alpha(),
            self.pixbuf.get_bits_per_sample(),
            width,
            height,
        )
        new_pixbuf.fill(0xFFFFFF00)

        self.pixbuf.copy_area(
            src_x=0,
            src_y=0,
            width=self.pixbuf.get_width(),
            height=self.pixbuf.get_height(),
            dest_pixbuf=new_pixbuf,
            dest_x=offset_x,
            dest_y=offset_y,
        )

        return ExtendedPixbuf(new_pixbuf)

    def copy(self):
        """
        Override to return same type
        """
        return ExtendedPixbuf(GdkPixbuf.Pixbuf.copy(self.pixbuf))

    def add_alpha(self, substitute_color, r, g, b):
        """
        Override to return same type
        """
        return ExtendedPixbuf(
            GdkPixbuf.Pixbuf.add_alpha(self.pixbuf, substitute_color, r, g, b)
        )

    def scale_simple(self, dest_width, dest_height, interp_type):
        """
        Override to return same type
        """
        return ExtendedPixbuf(
            GdkPixbuf.Pixbuf.scale_simple(
                self.pixbuf, dest_width, dest_height, interp_type
            )
        )

    def composite_color_simple(
        self,
        dest_width,
        dest_height,
        interp_type,
        overall_alpha,
        check_size,
        color1,
        color2,
    ):
        """
        Override to return same type
        """
        return ExtendedPixbuf(
            GdkPixbuf.Pixbuf.composite_color_simple(
                self.pixbuf,
                dest_width,
                dest_height,
                interp_type,
                overall_alpha,
                check_size,
                color1,
                color2,
            )
        )

    def new_subpixbuf(self, src_x, src_y, width, height):
        """
        Override to return same type
        """
        return ExtendedPixbuf(
            GdkPixbuf.Pixbuf.new_subpixbuf(self.pixbuf, src_x, src_y, width, height)
        )

    def rotate_simple(self, angle):
        """
        Override to return same type
        """
        return ExtendedPixbuf(GdkPixbuf.Pixbuf.rotate_simple(self.pixbuf, angle))

    def flip(self, horizontal):
        """
        Override to return sampe type
        """
        return ExtendedPixbuf(GdkPixbuf.Pixbuf.flip(self.pixbuf, horizontal))


def extended_pixbuf_new_from_file(filename):
    """
    Returns a new :class:`ExtendedPixbuf` containing
    an image loaded from the specified file

    :param filename: the name of the file
        containing the image to load
    :type filename: string
    :returns: a new pixbuf
    :rtype: :class:`ExtendedPixbuf`
    """
    return ExtendedPixbuf(GdkPixbuf.Pixbuf.new_from_file(filename))


class IconManager:
    """
    Provides convenience functions for
    managing icons and images in general
    """

    def __init__(self):
        self.icon_theme = Gtk.IconTheme.get_default()
        # TODO: Make svg actually recognized
        self._sizes = [16, 22, 24, 32, 48, 128, 'scalable']
        self._cache = {}

        self.rating_active_pixbuf = ExtendedPixbuf(
            self.pixbuf_from_icon_name('starred', size=Gtk.IconSize.MENU)
        )
        self.rating_inactive_pixbuf = ExtendedPixbuf(
            self.pixbuf_from_icon_name('non-starred', size=Gtk.IconSize.MENU)
        )

        # nobody actually sets the rating option, so don't handle it for now
        # event.add_ui_callback(self.on_option_set, 'rating_option_set')

    def add_icon_name_from_directory(
        self, icon_name: str, directory: str
    ) -> Optional[str]:
        """
        Registers an icon name from files found in a directory

        :param icon_name: the name for the icon
        :param directory: the location to search for icons
        :return: filesystem location of the highest-quality icon of this
            name, or None if not found
        """
        path = None
        for size in self._sizes:
            if isinstance(size, int):  # WxH/icon_name.png and scalable/icon_name.svg
                sizedir = '%dx%d' % (size, size)
            else:
                sizedir = size
            filepath = os.path.join(directory, sizedir, icon_name)
            files = glob.glob('%s.*' % filepath)
            if files:
                path = files[0]
                icon_size = size if size != 'scalable' else -1
                self.add_icon_name_from_file(icon_name, path, icon_size)
            else:  # icon_nameW.png and icon_name.svg
                if isinstance(size, int):
                    filename = '%s%d' % (icon_name, size)
                else:
                    filename = icon_name
                filepath = os.path.join(directory, filename)
                files = glob.glob('%s.*' % filepath)
                if files:
                    path = files[0]
                    icon_size = size if size != 'scalable' else -1
                    self.add_icon_name_from_file(icon_name, path, icon_size)
                else:  # Give up
                    pass
        return path

    def add_icon_name_from_file(
        self, icon_name: str, filename: str, size: Optional[int] = None
    ) -> None:
        """
        Registers an icon name from a filename

        :param icon_name: the name for the icon
        :type icon_name: string
        :param filename: the filename of an image
        :type filename: string
        :param size: the size the icon shall be registered for
        :type size: int
        """
        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
            self.add_icon_name_from_pixbuf(icon_name, pixbuf, size)
        except Exception as e:
            # Happens if, e.g., librsvg is not installed.
            logger.warning(
                'Failed to add icon name "{icon_name}" '
                'from file "{filename}": {error}'.format(
                    icon_name=icon_name, filename=filename, error=e.message
                )
            )

    def add_icon_name_from_pixbuf(
        self, icon_name: str, pixbuf: GdkPixbuf.Pixbuf, size: Optional[int] = None
    ) -> None:
        """
        Registers an icon name from a pixbuf

        :param icon_name: the name for the icon
        :param pixbuf: the pixbuf of an image
        :param size: the size the icon shall be registered for
        """
        if size is None:
            size = pixbuf.get_width()

        Gtk.IconTheme.add_builtin_icon(icon_name, size, pixbuf)

    def pixbuf_from_icon_name(
        self, icon_name: str, size: Union[int, Gtk.IconSize] = Gtk.IconSize.BUTTON
    ) -> Optional[GdkPixbuf.Pixbuf]:
        """
        Generates a pixbuf from an icon name

        :param icon_name: an icon name
        :param size: the size of the icon, will be
            tried to converted to a GTK icon size

        :returns: the generated pixbuf
        """
        if isinstance(size, Gtk.IconSize):
            icon_size = Gtk.icon_size_lookup(size)
            size = icon_size[1]

        if icon_name.endswith('-symbolic'):
            fallback_name = icon_name[:-9]
        else:
            fallback_name = icon_name + '-symbolic'

        icon_info = self.icon_theme.choose_icon(
            [icon_name, fallback_name],
            size,
            Gtk.IconLookupFlags.USE_BUILTIN | Gtk.IconLookupFlags.FORCE_SIZE,
        )
        if icon_info:
            try:
                return icon_info.load_icon()
            except GLib.Error as e:
                logger.warning('Failed to load icon "%s": %s', icon_name, e.message)
        else:
            logger.warning('Icon "%s" not found', icon_name)

        return None

    @common.cached(limit=settings.get_option('rating/maximum', 5) * 3)
    def pixbuf_from_rating(
        self, rating: int, size_ratio: float = 1
    ) -> GdkPixbuf.Pixbuf:
        """
        Returns a pixbuf representing a rating

        :param rating: the rating
        :returns: the rating pixbuf
        """
        maximum = settings.get_option('rating/maximum', 5)
        width = self.rating_active_pixbuf.get_width()
        height = self.rating_active_pixbuf.get_height()

        active_pixbuf = self.rating_active_pixbuf.scale_simple(
            int(width * size_ratio),
            int(height * size_ratio),
            GdkPixbuf.InterpType.BILINEAR,
        )
        inactive_pixbuf = self.rating_inactive_pixbuf.scale_simple(
            int(width * size_ratio),
            int(height * size_ratio),
            GdkPixbuf.InterpType.BILINEAR,
        )
        rating = max(0, rating)
        rating = min(rating, maximum)

        if rating == 0:
            return inactive_pixbuf * maximum
        elif rating == maximum:
            return active_pixbuf * maximum

        active_pixbufs = active_pixbuf * rating
        inactive_pixbufs = inactive_pixbuf * (maximum - rating)
        return active_pixbufs + inactive_pixbufs

    def __create_drag_cover_icon(self, pixbuf_list, cover_width):
        """
        Creates a stacked covers effect
        Put a transparent frame on it (look for `border`)
        :param pixbuf_list: list of GdkPixbuf.Pixbuf (covers)
        :param cover_width: int
        :return: GdkPixbuf.Pixbuf or None (if len(:param pixbuf_list:) == 0)
        """
        if len(pixbuf_list) == 0:
            return None

        # Sizes
        cover_side = cover_width
        cover_portion = int(cover_side * 0.08)  # 8% of sides
        covers_square_side = cover_side + cover_portion * (len(pixbuf_list) - 1)
        border = 4
        total_width = total_height = covers_square_side + (2 * border)

        # The result pixbuf
        result_pixbuf = GdkPixbuf.Pixbuf.new(
            GdkPixbuf.Colorspace.RGB, True, 8, total_width, total_height
        )
        result_pixbuf.fill(0x00000000)  # Fill with transparent background

        covers_area = result_pixbuf.new_subpixbuf(
            src_x=border,
            src_y=border,
            width=covers_square_side,
            height=covers_square_side,
        )
        # Write covers
        for i, pixbuf in enumerate(pixbuf_list):
            dest = cover_portion * i  # Put in a diagonal
            pixbuf.copy_area(
                0, 0, pixbuf.get_width(), pixbuf.get_height(), covers_area, dest, dest
            )

        return result_pixbuf

    def get_drag_cover_icon(self, tracks):
        """
        Get drag cover icon (stacked covers pixbuf)
        Asynchronous load covers for tracks taking at most 0.333 seconds
        :param tracks: iterable for tracks (xl.trax.Track)
        :return: GdkPixbuf.Pixbuf or None if none found
        """
        cover_width = settings.get_option('gui/cover_width', 100)
        as_pixbuf = lambda data: pixbuf_from_data(
            data=data, size=(cover_width, cover_width)
        )
        get_cover_for_tracks = covers.MANAGER.get_cover_for_tracks
        db_string_list = []
        cover_for_tracks = lambda tracks: get_cover_for_tracks(tracks, db_string_list)
        filtered_covers = filter(
            None, map(cover_for_tracks, tracks)
        )  # Remove None cover tracks
        async_loader = common.AsyncLoader(map(as_pixbuf, filtered_covers))
        async_loader.end(0.333)
        return self.__create_drag_cover_icon(async_loader.result, cover_width)


if 'EXAILE_BUILDING_DOCS' not in os.environ:
    MANAGER = IconManager()

# vim: et sts=4 sw=4



# Copyright (C) 2024  Johannes Sasongko <johannes sasongko org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


__all__ = ['SqliteDbm']


import operator
import os
import sqlite3
import sys
from typing import Iterator, Literal, MutableMapping, Union
import urllib.parse


class SqliteDbm(MutableMapping[bytes, bytes]):
    """
    DBM-compatible database implemented on top of SQLite.

    This class can operate on files created by dbm.sqlite (Python >= 3.13) and
    vice versa. The main difference is in how sqlite3.connect is called: this
    class uses check_same_thread=False and allows autocommit to be configured.
    """

    autocommit: bool
    conn: sqlite3.Connection

    def __init__(
        self,
        path: Union[str, bytes, os.PathLike],
        *,
        autocommit: bool = True,
        mode: Literal['ro', 'rw', 'rwc'] = 'rwc',
    ):
        """
        Open a database.

        :param path: Database file path.
        :param autocommit: Commit after every change instead of requiring sync.
        :param mode: Database open mode: ro = read only, rw = read/write, rwc =
            read/write/create.
        """

        self.autocommit = autocommit
        path_urlquoted = urllib.parse.quote(os.fsencode(path))
        uri = f"file:{path_urlquoted}?mode={mode}"
        if sys.version_info >= (3, 12):
            self.conn = conn = sqlite3.connect(
                uri, autocommit=autocommit, check_same_thread=False, uri=True
            )
        else:
            self.conn = conn = sqlite3.connect(
                uri,
                check_same_thread=False,
                isolation_level=None if autocommit else "DEFERRED",
                uri=True,
            )
        # Our database is a Dict table with (key, value) columns.
        # Python DBM interface uses bytes as the key and value type, which
        # corresponds to SQLite BLOB.
        # This doesn't use WITHOUT ROWID <https://www.sqlite.org/withoutrowid.html>
        # because our rows tend to be longer than the 200 B threshold mentioned.
        try:
            conn.execute(
                "CREATE TABLE IF NOT EXISTS Dict (key BLOB PRIMARY KEY NOT NULL, value BLOB NOT NULL)"
            )
        except Exception:
            conn.close()
            del self.conn
            raise

    def close(self) -> None:
        if hasattr(self, "conn"):
            try:
                self.sync()
            finally:
                self.conn.close()
                del self.conn

    def sync(self) -> None:
        if not self.autocommit:
            self.conn.commit()

    def __contains__(self, key: Union[str, bytes]) -> bool:
        key = self.__fix_type(key)
        return (
            self.conn.execute("SELECT NULL FROM Dict WHERE key = ?", (key,)).fetchone()
            is not None
        )

    __del__ = close

    def __delitem__(self, key: Union[str, bytes]) -> None:
        key = self.__fix_type(key)
        if self.conn.execute("DELETE FROM Dict WHERE key = ?", (key,)).rowcount == 0:
            raise KeyError(key)

    def __enter__(self) -> 'SqliteDbm':
        return self

    def __exit__(self, *_) -> None:
        self.close()

    def __getitem__(self, key: Union[str, bytes]) -> bytes:
        key = self.__fix_type(key)
        result = self.conn.execute(
            "SELECT value FROM Dict where key = ?", (key,)
        ).fetchone()
        if result is None:
            raise KeyError(key)
        return result[0]

    def __iter__(self) -> Iterator[bytes]:
        return map(operator.itemgetter(0), self.conn.execute("SELECT key FROM Dict"))

    def __len__(self) -> int:
        return self.conn.execute("SELECT COUNT(*) FROM Dict").fetchone()[0]

    def __setitem__(self, key: Union[str, bytes], value: Union[str, bytes]) -> None:
        key = self.__fix_type(key)
        value = self.__fix_type(value)
        self.conn.execute("REPLACE INTO Dict VALUES (?, ?)", (key, value))

    @staticmethod
    def __fix_type(obj: object) -> bytes:
        if isinstance(obj, bytes):
            return obj
        if isinstance(obj, str):
            return obj.encode('utf-8')
        raise TypeError(
            "key/value must be bytes or str (will be encoded to UTF-8 bytes)"
        )



"""
This module is a low-level reader and parser for audio CDs.
It heavily relies on ioctls to the linux kernel.

Original source for the code:
http://www.carey.geek.nz/code/python-cdrom/cdtoc.py

Source for all the magical constants and more infos on the ioctls:
linux/include/uapi/linux/cdrom.h
https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/include/uapi/linux/cdrom.h
"""

from __future__ import division


import fcntl
import logging
import os
import struct

from xl.trax import Track

logger = logging.getLogger(__name__)


def read_cd_index(device):
    """
    Reads a CD's index (table of contents, TOC).
    This must happen async because the I/O operations may take some time.

    @param device: a path to a CD device
    @return: Array of toc entries. The last one is a dummy. To be read by parse_tracks().
    """
    mcn = None
    toc_entries = []
    fd = os.open(device, os.O_RDONLY)
    try:
        start, end = __read_toc_header(fd)
        mcn = __read_disc_mcn(fd)

        # index of the end, i.e. the last toc entry which is an empty dummy
        CDROM_LEADOUT = 0xAA
        for toc_entry_index in list(range(start, end + 1)) + [CDROM_LEADOUT]:
            toc_entry = __read_toc_entry(fd, toc_entry_index)
            # XXX one could also reat+compute the `isrc` track id, see libdiscid
            toc_entries.append(toc_entry)
        logger.debug('Successfully read TOC of CD with MCN %s : %s', mcn, toc_entries)
    except Exception:
        logger.warn('Failed to read CD TOC', exc_info=True)
    finally:
        os.close(fd)
    # clear output for convenience
    if len(toc_entries) == 0:
        toc_entries = None
    return toc_entries, mcn


def __read_toc_header(fd):
    """A wrapper for the `CDROMREADTOCHDR` ioctl"""
    # struct cdrom_tochdr of 2 times u8
    FORMAT_cdrom_tochdr = 'BB'
    # u8 start: lowest track index (index of first track), probably always 1
    # u8 end: highest track index (index of last track), = number of tracks
    cdrom_tochdr = struct.pack(FORMAT_cdrom_tochdr, 0, 0)

    CDROMREADTOCHDR = 0x5305
    cdrom_tochdr = fcntl.ioctl(fd, CDROMREADTOCHDR, cdrom_tochdr)

    start, end = struct.unpack(FORMAT_cdrom_tochdr, cdrom_tochdr)

    return (start, end)


def __read_disc_mcn(fd):
    """A wrapper for the `CDROM_GET_MCN` ioctl"""
    # struct cdrom_mcn of 14 bytes, null-terminated
    FORMAT_cdrom_mcn = '14s'
    cdrom_mcn = struct.pack(FORMAT_cdrom_mcn, b'\0')

    CDROM_GET_MCN = 0x5311
    cdrom_mcn = fcntl.ioctl(fd, CDROM_GET_MCN, cdrom_mcn)

    raw_mcn = struct.unpack(FORMAT_cdrom_mcn, cdrom_mcn)
    mcn = raw_mcn[0][0:13]
    if b'0000000000000' in mcn:
        return None
    else:
        return mcn


def __read_toc_entry(fd, toc_entry_num):
    """A wrapper for the `CDROMREADTOCENTRY` ioctl"""
    # value constant: Minute, Second, Frame: binary (not bcd here)
    CDROM_MSF = 0x02

    # struct cdrom_tocentry of 3 times u8 followed by an int and another u8
    FORMAT_cdrom_tocentry = 'BBBiB'
    # u8 cdte_track: Track number. Starts with 1, which is used for the TOC and contains data.
    # u8 cdte_adr_ctrl: 4 high bits -> cdte_ctrl, 4 low bits -> cdte_adr
    # u8 cdte_format: should be CDROM_MSF=0x02 as requested before
    # int cdte_addr: see below
    # u8 cdte_datamode: ??? (ignored)
    cdrom_tocentry = struct.pack(
        FORMAT_cdrom_tocentry, toc_entry_num, 0, CDROM_MSF, 0, 0
    )

    CDROMREADTOCENTRY = 0x5306
    cdrom_tocentry = fcntl.ioctl(fd, CDROMREADTOCENTRY, cdrom_tocentry)

    cdte_track, cdte_adr_ctrl, cdte_format, cdte_addr, _cdte_datamode = struct.unpack(
        FORMAT_cdrom_tocentry, cdrom_tocentry
    )

    if cdte_format is not CDROM_MSF:
        raise OSError('Invalid syscall answer')

    # unused:
    # cdte_adr = cdte_adr_ctrl & 0x0f  # lower nibble

    cdte_ctrl = (cdte_adr_ctrl & 0xF0) >> 4  # higher nibble

    CDROM_DATA_TRACK = 0x04
    # data: `True` if this "track" contains data, `False` if it is audio
    is_data_track = bool(cdte_ctrl & CDROM_DATA_TRACK)

    # union cdrom_addr of struct cdrom_msf0 and int
    # struct cdrom_msf0 of 3 times u8 plus padding to match size of int
    FORMAT_cdrom_addr = 'BBB' + 'x' * (struct.calcsize('i') - 3)
    # u8 minute: Minutes from beginning of CD
    # u8 second: Seconds after `minute`
    # u8 frame: Frames after `frame`
    minute, second, frame = struct.unpack(
        FORMAT_cdrom_addr, struct.pack('i', cdte_addr)
    )

    return (cdte_track, is_data_track, minute, second, frame)


def parse_tracks(toc_entries, mcn, device):
    """
    Parses the given toc entries and mcn into tracks.
    As a result, the data will only contain track numbers and lengths but
    no sophisticated metadata.

    @param toc_entries: from read_cd_index()
    @param mcn: from read_cd_index()
    @param device: Name of the CD device
    @return: An array of xl.trax.Track with minimal information
    """
    real_track_count = len(toc_entries) - 1  # ignore the empty dummy track at the end
    tracks = []
    for toc_entry_index in range(0, real_track_count):
        track_index, is_data_track, _, _, _ = toc_entries[toc_entry_index]
        if is_data_track:
            continue
        if track_index is not toc_entry_index + 1:
            logger.warn(
                'Unexpected index found. %ith toc entry claims to be track number %i',
                toc_entry_index,
                track_index,
            )

        length = __calculate_track_length(
            toc_entries[toc_entry_index], toc_entries[toc_entry_index + 1]
        )
        track_uri = "cdda://%d/#%s" % (track_index, device)

        track = Track(uri=track_uri, scan=False)

        track_number = '{0}/{1}'.format(track_index, real_track_count)
        track.set_tags(
            title="Track %d" % track_index, tracknumber=track_number, __length=length
        )

        if mcn:
            track.set_tags(mcn=mcn)

        tracks.append(track)
    return tracks


def __calculate_track_length(current_track, next_track):
    """Calculate length of a single track from its data and the data of the following track"""
    _, _, begin_minute, begin_second, begin_frame = current_track
    _, _, end_minute, end_second, end_frame = next_track

    length_minutes = end_minute - begin_minute
    length_seconds = end_second - begin_second
    length_frames = end_frame - begin_frame
    # 75 frames per second, see CD_FRAMES in cdrom.h file
    length = length_minutes * 60 + length_seconds + length_frames / 75
    return length



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import xl.unicode
from xl.metadata._base import CaseInsensitiveBaseFormat, CoverImage
from xl import settings
from mutagen import flac
from mutagen.flac import Picture


class FlacFormat(CaseInsensitiveBaseFormat):
    MutagenType = flac.FLAC
    tag_mapping = {
        'cover': '__cover',
        'language': "Language",
        '__rating': 'rating',
    }
    writable = True
    case_sensitive = False

    def get_bitrate(self):
        return -1

    def get_keys_disk(self):
        keys = CaseInsensitiveBaseFormat.get_keys_disk(self)
        if self.mutagen.pictures:
            keys.append('cover')
        return keys

    def _get_tag(self, raw, tag):
        if tag == '__cover':
            return [
                CoverImage(type=p.type, desc=p.desc, mime=p.mime, data=p.data)
                for p in raw.pictures
            ]

        elif tag == 'rating':
            if 'rating' not in raw:
                return []
            # Rating Stars
            data = int(raw['rating'][0])
            return [str(self._rating_to_stars(data))]

        elif tag == 'bpm':
            if (
                settings.get_option('collection/use_legacy_metadata_mapping', False)
                and 'tempo' in raw
            ):
                tag = 'tempo'

        elif tag == 'comment':
            if (
                settings.get_option('collection/use_legacy_metadata_mapping', False)
                and 'description' in raw
            ):
                tag = 'description'

        return CaseInsensitiveBaseFormat._get_tag(self, raw, tag)

    def _set_tag(self, raw, tag, value):
        if tag == '__cover':
            raw.clear_pictures()
            for v in value:
                picture = Picture()
                picture.type = v.type
                picture.desc = v.desc
                picture.mime = v.mime
                picture.data = v.data
                raw.add_picture(picture)
            return

        elif tag == 'rating':
            # Rating Stars
            value = [str(self._stars_to_rating(int(value[0])))]

        elif tag == 'bpm':
            if settings.get_option('collection/use_legacy_metadata_mapping', False):
                tag = 'tempo'
            value = [xl.unicode.to_unicode(v) for v in value]

        elif tag == 'comment':
            if settings.get_option('collection/use_legacy_metadata_mapping', False):
                tag = 'description'
            value = [xl.unicode.to_unicode(v) for v in value]

        else:
            # flac has text based attributes, so convert everything to unicode
            value = [xl.unicode.to_unicode(v) for v in value]
        CaseInsensitiveBaseFormat._set_tag(self, raw, tag, value)

    def _del_tag(self, raw, tag):
        if tag == '__cover':
            raw.clear_pictures()
        elif tag in raw:
            del raw[tag]


# vim: et sts=4 sw=4



# Copyright (C) 2008 Erik Hetzner
#              2010 Brian Parma

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <https://www.gnu.org/licenses/>.

import spydaap.parser
from spydaap.daap import do
import logging

logger = logging.getLogger(__name__)


class ExaileParser(spydaap.parser.Parser):
    _string_map = {
        'title': 'dmap.itemname',
        'artist': 'daap.songartist',
        'composer': 'daap.songcomposer',
        'genre': 'daap.songgenre',
        'album': 'daap.songalbum',
    }

    _int_map = {
        'bpm': 'daap.songbeatsperminute',
        # not used by exaile client, and not parsed right anyway (ie: '2010-01-01')
        #        'date': 'daap.songyear', #TODO
        'year': 'daap.songyear',
        'tracknumber': 'daap.songtracknumber',
        'tracktotal': 'daap.songtrackcount',
        'discnumber': 'daap.songdiscnumber',
    }

    def understands(self, filename):
        return True

    #   return self.file_re.match(filename)

    # returns a list in exaile
    def handle_int_tags(self, map, md, daap):
        for k in md.list_tags():
            if k in map:
                try:
                    tn = str(md.get_tag_raw(k)[0])
                    if '/' in tn:
                        num, tot = tn.split('/')
                        if num == '':  # empty tags
                            num = 0
                        daap.append(do(map[k], int(num)))
                        # set total?
                    else:
                        daap.append(do(map[k], int(tn)))
                except Exception:
                    logger.exception(
                        'exception caught parsing tag: %s=%s from %s', k, tn, md
                    )

    # We can't use functions in __init__ because exaile tracks no longer
    # give us access to .tags
    def handle_string_tags(self, map, md, daap):
        for k in md.list_tags():
            if k in map:
                try:
                    tag = [t.encode("utf-8", "replace") for t in md.get_tag_raw(k)]
                    tag = [t for t in tag if t != ""]
                    daap.append(do(map[k], b"/".join(tag)))
                except Exception:
                    logger.exception("error decoding tags")

    def parse(self, trk):
        try:
            # trk = mutagen.File(filename)
            d = []
            if len(trk.list_tags()) > 0:
                if 'title' in trk.list_tags():
                    name = trk.get_tag_raw('title')[0].encode("utf-8", "replace")
                else:
                    name = str(trk)

                self.handle_string_tags(self._string_map, trk, d)
                self.handle_int_tags(self._int_map, trk, d)
            #                self.handle_rating(trk, d)
            else:
                name = str(trk)
            # statinfo = os.stat(filename)

            _len = trk.get_tag_raw('__length')
            if _len is None:  # don't parse songs that don't have length
                return (None, None)

            d.extend(
                [  # do('daap.songsize', trk.get_size()),
                    # do('daap.songdateadded', statinfo.st_ctime),
                    # do('daap.songdatemodified', statinfo.st_ctime),
                    do('daap.songtime', _len * 1000),
                    #                      do('daap.songbitrate', trk.get_tag_raw('__bitrate') / 1000),
                    #                      do('daap.songsamplerate', ogg.info.sample_rate), # todo ??
                    do(
                        'daap.songformat', trk.get_local_path().split('.')[-1]
                    ),  # todo ??
                    do('daap.songdescription', 'Exaile Streaming Audio'),
                ]
            )
            return (d, name)

        except Exception:
            logger.exception('caught exception while processing %s', trk)
            return (None, None)



# Bookmark plugin for Exaile media player
# Copyright (C) 2009-2011 Brian Parma
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import copy
import json
import logging
import os
import threading

from gi.repository import GLib
from gi.repository import Gtk

from xl import covers, player, trax, xdg, providers
from xl import common
from xl.nls import gettext as _
import xlgui
from xlgui.guiutil import pixbuf_from_data
from xlgui.widgets import dialogs, menu

LOGGER = logging.getLogger(__name__)


_smi = menu.simple_menu_item
_sep = menu.simple_separator


class Bookmark:
    """
    Manages a bookmark and provides a method to create a menu item.
    """

    __counter = 0

    def __init__(
        self, bookmarks_menu, delete_menu, delete_bookmark_callback, path, time
    ):
        """
        Creates a bookmark for current track/position if path or time are
        None. Creates a bookmark for the given track/position otherwise.
        """
        if not path:
            # get currently playing track
            track = player.PLAYER.current
            if track is None:
                text = 'Need a playing track to Bookmark.'
                LOGGER.error(text)
                dialogs.error(xlgui.main.mainwindow(), text)
                return
            time = player.PLAYER.get_time()
            path = track.get_loc_for_io()

        self.__path = path
        self.__time = time
        self.__title = None
        self.__cover_pixbuf = None
        self.item = None

        self.__fetch_metadata()
        self.__create_menu_item(bookmarks_menu, delete_menu, delete_bookmark_callback)

    def get_menu_item(self):
        return self.item

    def __fetch_metadata(self, *_args):
        # "gtk-menu-images" is deprecated and is being ignored on most
        # platforms. On GNOME/Wayland it can be enabled by editing
        # ~/.config/gtk-3.0/settings.ini and adding `gtk-menu-images=1`.
        use_covers = Gtk.Settings.get_default().props.gtk_menu_images

        try:
            item = trax.Track(self.__path)
            if not self.__title:
                self.__title = item.get_tag_display('title')
            if use_covers:
                image = covers.MANAGER.get_cover(item, set_only=True)
                if image:
                    try:
                        self.__cover_pixbuf = pixbuf_from_data(image, size=(16, 16))
                    except GLib.Error:
                        LOGGER.warning('Could not load cover')
            else:
                self.__cover_pixbuf = None
        except Exception:
            LOGGER.exception("Cannot open %s", self.__path)
            return

    def __create_menu_item(self, bookmarks_menu, delete_menu, delete_bookmark_callback):
        """
        Create menu entries for this bookmark.
        """
        time = '%d:%02d' % (self.__time // 60, self.__time % 60)
        label = '%s @ %s' % (self.__title, time)

        def factory(menu_, _parent, _context):
            "menu factory for new bookmarks"
            menu_item = Gtk.ImageMenuItem.new_with_mnemonic(label)
            if self.__cover_pixbuf:
                menu_item.set_image(Gtk.Image.new_from_pixbuf(self.__cover_pixbuf))

            if menu_ is bookmarks_menu:
                menu_item.connect('activate', self.__on_bookmark_activated)
            else:
                menu_item.connect('activate', delete_bookmark_callback, self)

            return menu_item

        item = menu.MenuItem('bookmark{0}'.format(Bookmark.__counter), factory, ['sep'])
        Bookmark.__counter += 1
        bookmarks_menu.add_item(item)
        delete_menu.add_item(item)
        self.item = item

    def __on_bookmark_activated(self, _widget):
        """
        This is called to resume a bookmark.
        """
        # check if it's already playing
        track = player.PLAYER.current
        if track and track.get_loc_for_io() == self.__path:
            player.PLAYER.unpause()
            player.PLAYER.seek(self.__time)
        else:
            # play it using the QUEUE
            track = trax.Track(self.__path)
            if track:  # make sure we got one
                player.QUEUE.play(track)
                player.PLAYER.seek(self.__time)

    def serialize_bookmark(self):
        """
        Serializes a bookmark object for json.dump().
        This function assumes that a bookmark argument can be used as if
        it was given as self. This assumption might be wrong in the future.
        """
        if not isinstance(self, Bookmark):
            raise ValueError()
        return (self.__path, self.__time)


class BookmarksManager:
    """
    Manages a list of bookmarks and the associated menu entries
    """

    __PATH = os.path.join(xdg.get_data_dirs()[0], 'bookmarklist.dat')

    def __init__(self):
        self.__db_file_lock = threading.RLock()
        self.__bookmarks = []
        # self.auto_db = {}

        self.menu = None
        self.delete_menu = None
        self.__setup_menu()

        # TODO: automatic bookmarks, not yet possible
        #  - needs a way to get the time a file is interrupted at
        # set events - not functional yet
        # event.add_callback(self.on_start_track, 'playback_start')
        # event.add_callback(self.on_stop_track, 'playback_end')
        # playback_end, playback_pause, playback_resume, stop_track

        self.__load_db()

    def __setup_menu(self):
        self.menu = menu.Menu(self)
        self.delete_menu = menu.Menu(self)

        def factory_factory(display_name, icon_name, callback=None, submenu=None):
            "define factory-factory for sensitive-aware menuitems"

            def factory(_menu, _parent, _context):
                item = Gtk.ImageMenuItem.new_with_mnemonic(display_name)
                image = Gtk.Image.new_from_icon_name(icon_name, size=Gtk.IconSize.MENU)
                item.set_image(image)

                if callback is not None:
                    item.connect('activate', callback)
                if submenu is not None:
                    item.set_submenu(submenu)
                    # insensitive if no bookmarks present
                    if len(self.__bookmarks) == 0:
                        item.set_sensitive(False)
                return item

            return factory

        items = []
        items.append(
            _smi(
                'bookmark',
                [],
                _('_Bookmark This Track'),
                'bookmark-new',
                self.__on_add_bookmark,
            )
        )
        delete_cb = factory_factory(
            _('_Delete Bookmark'), 'gtk-close', submenu=self.delete_menu
        )
        items.append(menu.MenuItem('delete', delete_cb, ['bookmark']))
        clear_cb = factory_factory(
            _('_Clear Bookmarks'), 'gtk-clear', callback=self.__clear_bookmarks
        )
        items.append(menu.MenuItem('clear', clear_cb, ['delete']))
        items.append(_sep('sep', ['clear']))

        for item in items:
            self.menu.add_item(item)

    def __on_add_bookmark(self, _widget, _name, _foo, _bookmarks_manager):
        self.__add_bookmark()

    def __add_bookmark(self, path=None, time=None, save_db=True):
        if not self.menu:
            return  # this plugin is shutting down
        bookmark = Bookmark(
            self.menu, self.delete_menu, self.__delete_bookmark, path, time
        )
        self.__bookmarks.append(bookmark)
        if save_db:
            self.__save_db()

    def __clear_bookmarks(self, _widget):
        """
        Delete all bookmarks.
        """
        for bookmark in self.__bookmarks:
            self.delete_menu.remove_item(bookmark.get_menu_item())
            self.menu.remove_item(bookmark.get_menu_item())
        self.__bookmarks = []
        self.__save_db()

    def __delete_bookmark(self, _widget, bookmark):
        """
        Delete a bookmark.
        """
        self.__bookmarks.remove(bookmark)
        self.delete_menu.remove_item(bookmark.get_menu_item())
        self.menu.remove_item(bookmark.get_menu_item())
        self.__save_db()

    @common.threaded
    def __load_db(self):
        """
        Load previously saved bookmarks from a file.
        """
        with self.__db_file_lock:
            if not os.path.exists(self.__PATH):
                LOGGER.info('Bookmarks file does not exist yet.')
                return
            try:
                # NOTE: both binary and non-binary works here; use
                # non-binary to be consistent with open() call in
                # __do_save_db(), which needs to be non-binary
                with open(self.__PATH, 'r') as bm_file:
                    bookmarks = json.load(bm_file)
                    self.__load_db_callback(bookmarks)
            except IOError as err:
                LOGGER.error('BM: could not open file: %s', err.strerror)

    @common.idle_add()
    def __load_db_callback(self, loaded_bookmarks):
        if not self.menu:
            return  # this plugin is shutting down

        for key, pos in loaded_bookmarks:
            self.__add_bookmark(key, pos, save_db=False)

    def __save_db(self):
        """
        Save list of bookmarks to a file.
        """
        # lists are not thread-safe, so we need a copy.
        # we don't need a deep copy because keys and values are not mutated.
        bookmarks = copy.copy(self.__bookmarks)
        # cannot use common.threaded here because it must not be daemonized,
        # otherwise we might loose data on program shutdown.
        thread = threading.Thread(target=self.__do_save_db, args=[bookmarks])
        thread.daemon = False
        thread.start()

    def __do_save_db(self, bookmarks):
        with self.__db_file_lock:
            # NOTE: intentionally open in non-binary mode because
            # otherwise json.dump() throws an error about trying to
            # write an str instead of bytes
            with open(self.__PATH, 'w') as bm_file:
                json.dump(
                    bookmarks, bm_file, indent=2, default=Bookmark.serialize_bookmark
                )
            LOGGER.debug('saved %d bookmarks', len(bookmarks))


class BookmarksPlugin:
    def __init__(self):
        self.__manager = None

    def enable(self, exaile):
        pass

    def on_gui_loaded(self):
        """
        Called when plugin is enabled.  Set up the menus, create the bookmark class, and
        load any saved bookmarks.
        """

        self.__manager = BookmarksManager()

        # add tools menu items
        providers.register(
            'menubar-tools-menu', _sep('plugin-sep', ['slow-scan-collection'])
        )
        item = _smi(
            'bookmarks',
            ['plugin-sep'],
            _('_Bookmarks'),
            'user-bookmarks',
            submenu=self.__manager.menu,
        )
        providers.register('menubar-tools-menu', item)

    def disable(self, _exaile):
        """
        Called when the plugin is disabled.  Destroy menu.
        """
        self.__manager.menu = None  # used to mark plugin shutdown to display_bookmark
        for item in providers.get('menubar-tools-menu'):
            if item.name == 'bookmarks':
                providers.unregister('menubar-tools-menu', item)
                break
        self.__manager = None


plugin_class = BookmarksPlugin

# vi: et ts=4 sts=4 sw=4



# -*- coding: utf-8  -*-

import logging

from xl.playlist import FormatConverter, PlaylistExportOptions

LOG = logging.getLogger(__name__)


class Test_FormatConverter:
    TIMEOUT = 2000
    MAX_ENTRIES = 2

    def test_get_track_export_path_relative(self):
        fc = FormatConverter('test')
        peo = PlaylistExportOptions
        peo.relative = True

        test = fc.get_track_export_path(
            'file:///tmp/playlist', 'file:///tmp/track1', peo
        )
        assert test == 'track1'

        test = fc.get_track_export_path(
            'file:///tmp/playlist', 'file:///tmp123/track1', peo
        )
        assert test == '../tmp123/track1'

        test = fc.get_track_export_path(
            'file:///tmp/playlist', 'file:///tmp123/track1%202', peo
        )
        assert test == '../tmp123/track1 2'

        test = fc.get_track_export_path(
            'gvfs:///tmp/playlist', 'gvfs:///tmp/track1', peo
        )
        assert test == 'track1'

        test = fc.get_track_export_path(
            'gvfs:///tmp/playlist', 'gvfs:///tmp123/track1', peo
        )
        assert test == '../tmp123/track1'

        test = fc.get_track_export_path(
            'gvfs:///tmp/playlist', 'http://ec3.yesstreaming.net:3420/stream', peo
        )
        assert test == 'http://ec3.yesstreaming.net:3420/stream'

        test = fc.get_track_export_path(
            'file:///tmp/playlist', 'http://ec3.yesstreaming.net:3420/stream', peo
        )
        assert test == 'http://ec3.yesstreaming.net:3420/stream'

    def test_track_export_path_stream(self):
        fc = FormatConverter('test')

        test = fc.get_track_export_path(
            'file:///tmp/playlist', 'http://ec3.yesstreaming.net:3420/stream', None
        )
        assert test == 'http://ec3.yesstreaming.net:3420/stream'

    def test_track_export_path_absolute(self):
        fc = FormatConverter('test')

        test = fc.get_track_export_path(
            'file:///tmp/playlist', 'file:///tmp/track1', None
        )
        assert test == '/tmp/track1'

        test = fc.get_track_export_path(
            'file:///tmp/playlist', 'file:///tmp123/track1%202', None
        )
        assert test == '/tmp123/track1 2'
